"""
spork.runtime.types - Core type definitions for Spork

This module contains the fundamental types used by both the runtime and compiler:
- Symbol: Represents symbolic identifiers in Spork code
- Keyword: Interned symbols that evaluate to themselves (like Clojure keywords)
- VectorLiteral: AST representation of vector literals [...]
- MapLiteral: AST representation of map literals {...}
- SetLiteral: AST representation of set literals #{...}
- Decorated: Represents decorator expressions (^decorator or ^(decorator args))
- MatchError: Exception raised when pattern matching fails
- normalize_name: Converts Lisp-style names to valid Python identifiers

These types are used by the reader/parser to represent Spork forms,
and by the runtime for values that exist at execution time.
"""

from dataclasses import dataclass
from typing import Any

# Sentinel for missing values
_MISSING = object()


class MatchError(Exception):
    """Raised when no pattern matches in a match expression or multi-dispatch function."""

    pass


def normalize_name(name: str) -> str:
    """
    Normalize a Lisp-style identifier to a valid Python identifier.

    Converts hyphens to underscores and special characters to safe suffixes.
    This is the single source of truth used by both the compiler (codegen)
    and the runtime (setup_runtime_env).
    """
    # Handle special operator names that are valid Spork symbols
    # but not valid Python identifiers
    SPECIAL_NAMES = {
        "+": "_plus_",
        "-": "_minus_",
        "*": "_star_",
        "/": "_slash_",
        "=": "_eq_",
        "<": "_lt_",
        ">": "_gt_",
        "<=": "_lte_",
        ">=": "_gte_",
        "!=": "_neq_",
        "==": "_eqeq_",
    }

    # Check for exact match first (operators used as values)
    if name in SPECIAL_NAMES:
        return SPECIAL_NAMES[name]

    # Replace hyphens with underscores
    result = name.replace("-", "_")

    # Replace trailing ? with _q (predicates like even?, nil?)
    if result.endswith("?"):
        result = result[:-1] + "_q"

    # Replace trailing ! with _bang (mutating functions like set!, swap!)
    if result.endswith("!"):
        result = result[:-1] + "_bang"

    # Replace other special characters that might appear mid-name
    result = result.replace("?", "_q_")
    result = result.replace("!", "_bang_")
    result = result.replace("*", "_star_")
    result = result.replace("+", "_plus_")
    result = result.replace("'", "_prime_")
    result = result.replace("$", "_S_")

    return result


@dataclass
class Symbol:
    """
    Represents a symbolic identifier in Spork code.

    Symbols are the fundamental naming construct in Spork. They represent
    variable names, function names, and other identifiers.

    Attributes:
        name: The string name of the symbol
        line: Source line number (1-based)
        col: Source column number (0-based)
        end_line: Ending line number
        end_col: Ending column number
    """

    name: str
    line: int = 0
    col: int = 0
    end_line: int = 0
    end_col: int = 0

    def __repr__(self):
        return self.name

    def with_location(
        self, line: int, col: int, end_line: int = 0, end_col: int = 0
    ) -> "Symbol":
        """Return a new Symbol with the given location."""
        return Symbol(
            self.name, line, col, end_line or line, end_col or col + len(self.name)
        )


@dataclass(eq=False)
class Keyword:
    """
    Keyword type - like Clojure keywords, these are interned symbols that
    evaluate to themselves and can be used as map keys.

    Keywords compare equal by name only (source location is ignored).
    They are hashable and can be used as dictionary keys.
    Keywords are also callable - calling a keyword with a map returns the value
    at that key: (:foo {:foo 1}) => 1

    Attributes:
        name: The string name of the keyword (without the leading colon)
        line: Source line number (1-based)
        col: Source column number (0-based)
        end_line: Ending line number
        end_col: Ending column number
    """

    name: str
    line: int = 0
    col: int = 0
    end_line: int = 0
    end_col: int = 0

    def __repr__(self):
        return f":{self.name}"

    def __str__(self):
        return f":{self.name}"

    def __eq__(self, other):
        if isinstance(other, Keyword):
            return self.name == other.name
        return False

    def __hash__(self):
        return hash(("Keyword", self.name))

    def __call__(self, coll, default=None):
        """
        Make keywords callable for map lookup.
        (:name {:name "Alice"}) => "Alice"
        (:missing {:name "Alice"} "default") => "default"
        """
        try:
            # Try direct key access first (works for Map and dict)
            if hasattr(coll, "get"):
                return coll.get(self, default)
            # Fallback to indexing
            return coll[self] if self in coll else default
        except (KeyError, TypeError):
            return default

    def with_location(
        self, line: int, col: int, end_line: int = 0, end_col: int = 0
    ) -> "Keyword":
        """Return a new Keyword with the given location."""
        return Keyword(
            self.name, line, col, end_line or line, end_col or col + len(self.name) + 1
        )


@dataclass
class Decorated:
    """
    Represents a decorator expression (^decorator or ^(decorator args)).

    Used in Spork for type annotations and metadata on definitions.

    Attributes:
        expr: The decorator expression (could be a Symbol, list, etc.)
        line: Source line number
        col: Source column number
        end_line: Ending line number
        end_col: Ending column number
    """

    expr: Any
    line: int = 0
    col: int = 0
    end_line: int = 0
    end_col: int = 0

    def __repr__(self):
        return f"Decorated({self.expr!r})"

    def with_location(
        self, line: int, col: int, end_line: int = 0, end_col: int = 0
    ) -> "Decorated":
        """Return a new Decorated with the given location."""
        return Decorated(self.expr, line, col, end_line or line, end_col)


@dataclass
class VectorLiteral:
    """
    AST node representing a vector literal [...] in source code.

    This is distinct from the runtime Vector type (pds.Vector).
    VectorLiteral is used during parsing and compilation, while Vector
    is the persistent data structure used at runtime.

    Attributes:
        items: List of elements in the vector
        line: Source line number
        col: Source column number
        end_line: Ending line number
        end_col: Ending column number
    """

    items: list[Any]
    line: int = 0
    col: int = 0
    end_line: int = 0
    end_col: int = 0

    def __repr__(self):
        return f"VectorLiteral({self.items!r})"

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def __getitem__(self, index):
        return self.items[index]

    def with_location(
        self, line: int, col: int, end_line: int = 0, end_col: int = 0
    ) -> "VectorLiteral":
        """Return a new VectorLiteral with the given location."""
        return VectorLiteral(self.items, line, col, end_line or line, end_col)


@dataclass
class MapLiteral:
    """
    Represents a map/dict literal that preserves the order and structure of key-value pairs.

    This is used instead of Python dict to support Clojure-style destructuring patterns
    like {a :x b :y} where the binding symbol comes before the key.

    Attributes:
        pairs: List of (key, value) tuples in the order they appeared
        line: Source line number
        col: Source column number
        end_line: Ending line number
        end_col: Ending column number
    """

    pairs: list[tuple[Any, Any]]
    line: int = 0
    col: int = 0
    end_line: int = 0
    end_col: int = 0

    def __repr__(self):
        return f"MapLiteral({self.pairs!r})"

    def with_location(
        self, line: int, col: int, end_line: int = 0, end_col: int = 0
    ) -> "MapLiteral":
        """Return a new MapLiteral with the given location."""
        return MapLiteral(self.pairs, line, col, end_line or line, end_col)


@dataclass
class SetLiteral:
    """
    AST node representing a set literal #{...} in source code.

    This is distinct from the runtime Set type (pds.Set).

    Attributes:
        items: List of elements in the set
        line: Source line number
        col: Source column number
        end_line: Ending line number
        end_col: Ending column number
    """

    items: list[Any]
    line: int = 0
    col: int = 0
    end_line: int = 0
    end_col: int = 0

    def __repr__(self):
        return f"SetLiteral({self.items!r})"

    def with_location(
        self, line: int, col: int, end_line: int = 0, end_col: int = 0
    ) -> "SetLiteral":
        """Return a new SetLiteral with the given location."""
        return SetLiteral(self.items, line, col, end_line or line, end_col)


@dataclass
class KwargsLiteral:
    """
    AST node representing a keyword arguments splat *{...} in source code.

    This is used in function calls to pass keyword arguments:
        (f 1 2 *{:name "Alice" :age 30})  => f(1, 2, name="Alice", age=30)

    Attributes:
        pairs: List of (key, value) tuples in the order they appeared
        line: Source line number
        col: Source column number
        end_line: Ending line number
        end_col: Ending column number
    """

    pairs: list[tuple[Any, Any]]
    line: int = 0
    col: int = 0
    end_line: int = 0
    end_col: int = 0

    def __repr__(self):
        return f"KwargsLiteral({self.pairs!r})"

    def with_location(
        self, line: int, col: int, end_line: int = 0, end_col: int = 0
    ) -> "KwargsLiteral":
        """Return a new KwargsLiteral with the given location."""
        return KwargsLiteral(self.pairs, line, col, end_line or line, end_col)


# Type exports
__all__ = [
    "Symbol",
    "Keyword",
    "Decorated",
    "VectorLiteral",
    "MapLiteral",
    "SetLiteral",
    "KwargsLiteral",
    "MatchError",
    "_MISSING",
]



"""
spork.project.config - Project configuration loader

This module handles parsing and loading spork.it project manifest files.
It provides the ProjectConfig class which holds all project metadata
and configuration.

The spork.it file uses Spork map syntax:
    {:name "my-project"
     :version "0.1.0"
     :description "A sample project"
     :dependencies ["requests" "numpy>=1.20"]
     :source-paths ["src"]
     :test-paths ["tests"]
     :main "my-project.core/main"}
"""

import os
from dataclasses import dataclass, field
from typing import Any, Optional

from spork.compiler.reader import read_str
from spork.runtime.types import Keyword, MapLiteral, VectorLiteral

# Default configuration values
DEFAULT_SOURCE_PATHS = ["src"]
DEFAULT_TEST_PATHS = ["tests"]
PROJECT_FILENAME = "spork.it"


def spork_to_python(value: Any) -> Any:
    """
    Convert Spork types to Python native types for internal tooling use.

    - Keyword -> str (without the colon)
    - VectorLiteral -> list
    - MapLiteral -> dict
    - Other types pass through unchanged
    """
    if isinstance(value, Keyword):
        return value.name
    elif isinstance(value, VectorLiteral):
        return [spork_to_python(item) for item in value.items]
    elif isinstance(value, MapLiteral):
        return {spork_to_python(k): spork_to_python(v) for k, v in value.pairs}
    elif isinstance(value, list):
        return [spork_to_python(item) for item in value]
    elif isinstance(value, dict):
        return {spork_to_python(k): spork_to_python(v) for k, v in value.items()}
    else:
        return value


def find_project_root(start_path: Optional[str] = None) -> Optional[str]:
    """
    Find the project root by walking up directory trees looking for spork.it.

    Args:
        start_path: Path to start searching from. If None, uses current working directory.
                   Can be a file or directory path.

    Returns:
        Absolute path to the directory containing spork.it, or None if not found.
    """
    if start_path is None:
        current = os.getcwd()
    elif os.path.isfile(start_path):
        current = os.path.dirname(os.path.abspath(start_path))
    else:
        current = os.path.abspath(start_path)

    # Walk up the directory tree
    while True:
        project_file = os.path.join(current, PROJECT_FILENAME)
        if os.path.isfile(project_file):
            return current

        parent = os.path.dirname(current)
        if parent == current:
            # Reached filesystem root
            return None
        current = parent


@dataclass
class ProjectConfig:
    """
    Represents a Spork project configuration loaded from spork.it.

    Required fields:
        name: Project name (string)
        version: Project version (string, e.g., "0.1.0")

    Optional fields:
        description: Project description
        dependencies: List of dependency specifications (pip-style)
        source_paths: List of source directories (default: ["src"])
        test_paths: List of test directories (default: ["tests"])
        main: Entry point function (e.g., "my-app.core/main")

    Computed fields:
        project_root: Absolute path to the directory containing spork.it
    """

    name: str
    version: str
    project_root: str
    description: Optional[str] = None
    dependencies: list[str] = field(default_factory=list)
    source_paths: list[str] = field(default_factory=lambda: DEFAULT_SOURCE_PATHS.copy())
    test_paths: list[str] = field(default_factory=lambda: DEFAULT_TEST_PATHS.copy())
    main: Optional[str] = None

    # Store the raw config for any additional fields
    _raw: dict[str, Any] = field(default_factory=dict, repr=False)

    @property
    def venv_path(self) -> str:
        """Path to the project's virtual environment."""
        return os.path.join(self.project_root, ".venv")

    @property
    def venv_python(self) -> str:
        """Path to the Python executable in the venv."""
        if os.name == "nt":  # Windows
            return os.path.join(self.venv_path, "Scripts", "python.exe")
        else:
            return os.path.join(self.venv_path, "bin", "python")

    @property
    def venv_pip(self) -> str:
        """Path to the pip executable in the venv."""
        if os.name == "nt":  # Windows
            return os.path.join(self.venv_path, "Scripts", "pip.exe")
        else:
            return os.path.join(self.venv_path, "bin", "pip")

    @property
    def venv_site_packages(self) -> Optional[str]:
        """Path to the site-packages directory in the venv."""
        if os.name == "nt":
            site_packages = os.path.join(self.venv_path, "Lib", "site-packages")
        else:
            # Find the Python version directory
            lib_path = os.path.join(self.venv_path, "lib")
            if os.path.isdir(lib_path):
                for entry in os.listdir(lib_path):
                    if entry.startswith("python"):
                        site_packages = os.path.join(lib_path, entry, "site-packages")
                        if os.path.isdir(site_packages):
                            return site_packages
            return None
        return site_packages if os.path.isdir(site_packages) else None

    def get_absolute_source_paths(self) -> list[str]:
        """Return absolute paths for all source directories."""
        return [os.path.join(self.project_root, p) for p in self.source_paths]

    def get_absolute_test_paths(self) -> list[str]:
        """Return absolute paths for all test directories."""
        return [os.path.join(self.project_root, p) for p in self.test_paths]

    def has_venv(self) -> bool:
        """Check if the virtual environment exists."""
        return os.path.isdir(self.venv_path) and os.path.isfile(self.venv_python)

    @classmethod
    def load(cls, path: Optional[str] = None) -> "ProjectConfig":
        """
        Load a ProjectConfig from a spork.it file.

        Args:
            path: Path to spork.it file, directory containing it, or None to search
                  from current directory upward.

        Returns:
            Loaded ProjectConfig instance.

        Raises:
            FileNotFoundError: If no spork.it file can be found.
            ValueError: If the spork.it file is invalid or missing required fields.
        """
        # Determine the project file path
        if path is None:
            project_root = find_project_root()
            if project_root is None:
                raise FileNotFoundError(
                    f"Could not find {PROJECT_FILENAME} in current directory or any parent directory"
                )
        elif os.path.isfile(path):
            if os.path.basename(path) == PROJECT_FILENAME:
                project_root = os.path.dirname(os.path.abspath(path))
            else:
                # Assume it's a file within a project, search upward
                project_root = find_project_root(path)
                if project_root is None:
                    raise FileNotFoundError(
                        f"Could not find {PROJECT_FILENAME} starting from {path}"
                    )
        elif os.path.isdir(path):
            project_file = os.path.join(path, PROJECT_FILENAME)
            if os.path.isfile(project_file):
                project_root = os.path.abspath(path)
            else:
                # Search upward from this directory
                project_root = find_project_root(path)
                if project_root is None:
                    raise FileNotFoundError(
                        f"Could not find {PROJECT_FILENAME} in {path} or any parent directory"
                    )
        else:
            raise FileNotFoundError(f"Path does not exist: {path}")

        project_file = os.path.join(project_root, PROJECT_FILENAME)

        # Read and parse the file
        with open(project_file, encoding="utf-8") as f:
            content = f.read()

        try:
            parsed = read_str(content)
        except Exception as e:
            raise ValueError(f"Failed to parse {project_file}: {e}") from e

        # read_str returns a list of all forms in the file
        # We need to find the first map (the config)
        if isinstance(parsed, list):
            # Find the first MapLiteral in the list
            config_form = None
            for form in parsed:
                if isinstance(form, MapLiteral):
                    config_form = form
                    break
            if config_form is None:
                raise ValueError(f"{project_file} must contain a map as the main form")
            parsed = config_form

        # Convert to Python types
        config_dict = spork_to_python(parsed)

        if not isinstance(config_dict, dict):
            raise ValueError(
                f"{project_file} must contain a map, got {type(parsed).__name__}"
            )

        # Validate required fields
        if "name" not in config_dict:
            raise ValueError(f"{project_file} is missing required field :name")
        if "version" not in config_dict:
            raise ValueError(f"{project_file} is missing required field :version")

        # Extract fields with defaults
        name = config_dict["name"]
        version = config_dict["version"]
        description = config_dict.get("description")
        dependencies = config_dict.get("dependencies", [])
        source_paths = config_dict.get("source-paths", DEFAULT_SOURCE_PATHS.copy())
        test_paths = config_dict.get("test-paths", DEFAULT_TEST_PATHS.copy())
        main = config_dict.get("main")

        # Validate types
        if not isinstance(name, str):
            raise ValueError(f":name must be a string, got {type(name).__name__}")
        if not isinstance(version, str):
            raise ValueError(f":version must be a string, got {type(version).__name__}")
        if description is not None and not isinstance(description, str):
            raise ValueError(
                f":description must be a string, got {type(description).__name__}"
            )
        if not isinstance(dependencies, list):
            raise ValueError(
                f":dependencies must be a vector, got {type(dependencies).__name__}"
            )
        if not isinstance(source_paths, list):
            raise ValueError(
                f":source-paths must be a vector, got {type(source_paths).__name__}"
            )
        if not isinstance(test_paths, list):
            raise ValueError(
                f":test-paths must be a vector, got {type(test_paths).__name__}"
            )
        if main is not None and not isinstance(main, str):
            raise ValueError(f":main must be a string, got {type(main).__name__}")

        return cls(
            name=name,
            version=version,
            project_root=project_root,
            description=description,
            dependencies=dependencies,
            source_paths=source_paths,
            test_paths=test_paths,
            main=main,
            _raw=config_dict,
        )


def load_config(path: Optional[str] = None) -> ProjectConfig:
    """
    Convenience function to load a ProjectConfig.

    Args:
        path: Path to spork.it file, directory containing it, or None to search
              from current directory upward.

    Returns:
        Loaded ProjectConfig instance.
    """
    return ProjectConfig.load(path)



"""
Test suite for the Spork Language Server Protocol implementation.

This module tests the LSP server functionality including:
- Protocol message framing
- Initialize/shutdown lifecycle
- Document synchronization
- Completion
- Hover
- Go to definition
- Diagnostics
"""

import io
import json
import unittest
from typing import Any


class TestProtocol(unittest.TestCase):
    """Test the JSON-RPC protocol implementation."""

    def test_message_framing(self):
        """Test that messages are properly framed with Content-Length headers."""
        from spork.lsp.protocol import ProtocolReader, ProtocolWriter

        # Create a buffer to write to
        output = io.BytesIO()
        writer = ProtocolWriter(output)

        # Write a message
        message = {"jsonrpc": "2.0", "method": "test", "params": {}}
        writer.write_message(message)

        # Read it back
        output.seek(0)
        reader = ProtocolReader(output)
        received = reader.read_message()

        self.assertEqual(received, message)

    def test_multiple_messages(self):
        """Test reading multiple messages in sequence."""
        from spork.lsp.protocol import ProtocolReader, ProtocolWriter

        output = io.BytesIO()
        writer = ProtocolWriter(output)

        messages = [
            {"jsonrpc": "2.0", "id": 1, "method": "first"},
            {"jsonrpc": "2.0", "id": 2, "method": "second"},
            {"jsonrpc": "2.0", "id": 3, "method": "third"},
        ]

        for msg in messages:
            writer.write_message(msg)

        output.seek(0)
        reader = ProtocolReader(output)

        for expected in messages:
            received = reader.read_message()
            self.assertEqual(received, expected)

    def test_unicode_content(self):
        """Test that unicode content is handled correctly."""
        from spork.lsp.protocol import ProtocolReader, ProtocolWriter

        output = io.BytesIO()
        writer = ProtocolWriter(output)

        message = {
            "jsonrpc": "2.0",
            "method": "test",
            "params": {"text": "こんにちは世界"},
        }
        writer.write_message(message)

        output.seek(0)
        reader = ProtocolReader(output)
        received = reader.read_message()

        self.assertEqual(received["params"]["text"], "こんにちは世界")


class TestJsonRpcProtocol(unittest.TestCase):
    """Test the JSON-RPC protocol handler."""

    def test_request_handling(self):
        """Test that requests are dispatched to handlers."""
        from spork.lsp.protocol import JsonRpcProtocol

        protocol = JsonRpcProtocol()

        # Register a handler
        def handle_test(params):
            return {"result": params.get("value", 0) * 2}

        protocol.register_request_handler("test", handle_test)

        # Handle a request
        message = {"jsonrpc": "2.0", "id": 1, "method": "test", "params": {"value": 21}}
        response = protocol.handle_message(message)

        self.assertEqual(response["id"], 1)
        self.assertEqual(response["result"]["result"], 42)

    def test_notification_handling(self):
        """Test that notifications don't produce responses."""
        from spork.lsp.protocol import JsonRpcProtocol

        protocol = JsonRpcProtocol()

        received = []

        def handle_notify(params):
            received.append(params)

        protocol.register_notification_handler("notify", handle_notify)

        # Handle a notification (no id)
        message = {"jsonrpc": "2.0", "method": "notify", "params": {"data": "test"}}
        response = protocol.handle_message(message)

        self.assertIsNone(response)
        self.assertEqual(len(received), 1)
        self.assertEqual(received[0]["data"], "test")

    def test_method_not_found(self):
        """Test that unknown methods return an error."""
        from spork.lsp.protocol import ErrorCode, JsonRpcProtocol

        protocol = JsonRpcProtocol()

        message = {"jsonrpc": "2.0", "id": 1, "method": "unknown"}
        response = protocol.handle_message(message)

        self.assertEqual(response["id"], 1)
        self.assertIn("error", response)
        self.assertEqual(response["error"]["code"], ErrorCode.METHOD_NOT_FOUND)


class TestLspHelpers(unittest.TestCase):
    """Test LSP helper functions."""

    def test_uri_to_path(self):
        """Test file URI to path conversion."""
        from spork.lsp.protocol import uri_to_path

        # Unix path
        self.assertEqual(
            uri_to_path("file:///home/user/test.spork"), "/home/user/test.spork"
        )

        # Path with spaces
        self.assertEqual(
            uri_to_path("file:///home/user/my%20project/test.spork"),
            "/home/user/my%20project/test.spork",
        )

    def test_path_to_uri(self):
        """Test path to file URI conversion."""
        from spork.lsp.protocol import path_to_uri

        # The result should start with file://
        uri = path_to_uri("/home/user/test.spork")
        self.assertTrue(uri.startswith("file://"))
        self.assertIn("test.spork", uri)

    def test_make_position(self):
        """Test position creation."""
        from spork.lsp.protocol import make_position

        pos = make_position(10, 5)
        self.assertEqual(pos["line"], 10)
        self.assertEqual(pos["character"], 5)

    def test_make_range(self):
        """Test range creation."""
        from spork.lsp.protocol import make_range

        range_ = make_range(1, 0, 1, 10)
        self.assertEqual(range_["start"]["line"], 1)
        self.assertEqual(range_["start"]["character"], 0)
        self.assertEqual(range_["end"]["line"], 1)
        self.assertEqual(range_["end"]["character"], 10)

    def test_make_diagnostic(self):
        """Test diagnostic creation."""
        from spork.lsp.protocol import DiagnosticSeverity, make_diagnostic, make_range

        range_ = make_range(0, 0, 0, 5)
        diag = make_diagnostic(
            range_=range_,
            message="Test error",
            severity=DiagnosticSeverity.ERROR,
            source="spork",
            code="E001",
        )

        self.assertEqual(diag["message"], "Test error")
        self.assertEqual(diag["severity"], DiagnosticSeverity.ERROR)
        self.assertEqual(diag["source"], "spork")
        self.assertEqual(diag["code"], "E001")

    def test_make_completion_item(self):
        """Test completion item creation."""
        from spork.lsp.protocol import CompletionItemKind, make_completion_item

        item = make_completion_item(
            label="defn",
            kind=CompletionItemKind.KEYWORD,
            detail="Define a function",
            documentation="Defines a named function.",
        )

        self.assertEqual(item["label"], "defn")
        self.assertEqual(item["kind"], CompletionItemKind.KEYWORD)
        self.assertEqual(item["detail"], "Define a function")
        self.assertEqual(item["documentation"], "Defines a named function.")

    def test_make_hover(self):
        """Test hover creation."""
        from spork.lsp.protocol import make_hover

        hover = make_hover("**defn** (macro)\n\nDefines a function.")

        self.assertEqual(hover["contents"]["kind"], "markdown")
        self.assertIn("defn", hover["contents"]["value"])


class TestTextDocument(unittest.TestCase):
    """Test the TextDocument class."""

    def test_get_line(self):
        """Test getting a specific line from a document."""
        from spork.lsp.server import TextDocument

        doc = TextDocument(
            uri="file:///test.spork",
            language_id="spork",
            version=1,
            content="line one\nline two\nline three",
        )

        self.assertEqual(doc.get_line(0), "line one")
        self.assertEqual(doc.get_line(1), "line two")
        self.assertEqual(doc.get_line(2), "line three")
        self.assertEqual(doc.get_line(99), "")

    def test_get_word_at_position(self):
        """Test extracting the word at a given position."""
        from spork.lsp.server import TextDocument

        doc = TextDocument(
            uri="file:///test.spork",
            language_id="spork",
            version=1,
            content="(defn hello-world [x] (+ x 1))",
        )

        # At 'defn'
        self.assertEqual(doc.get_word_at_position(0, 2), "defn")
        self.assertEqual(doc.get_word_at_position(0, 5), "defn")

        # At 'hello-world'
        self.assertEqual(doc.get_word_at_position(0, 10), "hello-world")

        # At 'x'
        self.assertEqual(doc.get_word_at_position(0, 19), "x")

    def test_path_property(self):
        """Test the path property."""
        from spork.lsp.server import TextDocument

        doc = TextDocument(
            uri="file:///home/user/test.spork",
            language_id="spork",
            version=1,
            content="",
        )

        self.assertEqual(doc.path, "/home/user/test.spork")


class TestSporkLanguageServer(unittest.TestCase):
    """Test the Spork Language Server."""

    def create_server(self):
        """Create a server instance for testing."""
        from spork.lsp.protocol import JsonRpcProtocol, ProtocolReader, ProtocolWriter
        from spork.lsp.server import SporkLanguageServer

        # Use BytesIO for input/output
        input_buffer = io.BytesIO()
        output_buffer = io.BytesIO()

        protocol = JsonRpcProtocol(
            reader=ProtocolReader(input_buffer),
            writer=ProtocolWriter(output_buffer),
        )

        server = SporkLanguageServer(protocol=protocol)
        return server, input_buffer, output_buffer

    def test_initialize(self):
        """Test the initialize request."""
        server, _, _ = self.create_server()

        params = {
            "processId": 1234,
            "rootUri": "file:///tmp/test",
            "capabilities": {},
        }

        result = server._handle_initialize(params)

        self.assertIn("capabilities", result)
        self.assertIn("textDocumentSync", result["capabilities"])
        self.assertIn("completionProvider", result["capabilities"])
        self.assertIn("hoverProvider", result["capabilities"])
        self.assertIn("definitionProvider", result["capabilities"])

    def test_did_open(self):
        """Test the textDocument/didOpen notification."""
        server, _, _ = self.create_server()

        # Initialize first
        server._handle_initialize({"processId": 1234, "capabilities": {}})
        server._init_backend()

        params = {
            "textDocument": {
                "uri": "file:///test.spork",
                "languageId": "spork",
                "version": 1,
                "text": "(defn add [a b] (+ a b))",
            }
        }

        server._handle_did_open(params)

        self.assertIn("file:///test.spork", server.documents)
        doc = server.documents["file:///test.spork"]
        self.assertEqual(doc.version, 1)
        self.assertIn("defn", doc.content)

    def test_did_change(self):
        """Test the textDocument/didChange notification."""
        server, _, _ = self.create_server()

        # Initialize and open a document
        server._handle_initialize({"processId": 1234, "capabilities": {}})
        server._init_backend()

        server._handle_did_open(
            {
                "textDocument": {
                    "uri": "file:///test.spork",
                    "languageId": "spork",
                    "version": 1,
                    "text": "(def x 1)",
                }
            }
        )

        # Change the document
        server._handle_did_change(
            {
                "textDocument": {"uri": "file:///test.spork", "version": 2},
                "contentChanges": [{"text": "(def x 2)"}],
            }
        )

        doc = server.documents["file:///test.spork"]
        self.assertEqual(doc.version, 2)
        self.assertIn("2", doc.content)

    def test_did_close(self):
        """Test the textDocument/didClose notification."""
        server, _, _ = self.create_server()

        server._handle_initialize({"processId": 1234, "capabilities": {}})
        server._init_backend()

        server._handle_did_open(
            {
                "textDocument": {
                    "uri": "file:///test.spork",
                    "languageId": "spork",
                    "version": 1,
                    "text": "(def x 1)",
                }
            }
        )

        self.assertIn("file:///test.spork", server.documents)

        server._handle_did_close({"textDocument": {"uri": "file:///test.spork"}})

        self.assertNotIn("file:///test.spork", server.documents)

    def test_completion_keywords(self):
        """Test that completion returns keywords."""
        server, _, _ = self.create_server()

        server._handle_initialize({"processId": 1234, "capabilities": {}})
        server._init_backend()

        server._handle_did_open(
            {
                "textDocument": {
                    "uri": "file:///test.spork",
                    "languageId": "spork",
                    "version": 1,
                    "text": "(def",
                }
            }
        )

        result = server._handle_completion(
            {
                "textDocument": {"uri": "file:///test.spork"},
                "position": {"line": 0, "character": 4},
            }
        )

        self.assertIn("items", result)
        labels = [item["label"] for item in result["items"]]
        self.assertIn("defn", labels)
        self.assertIn("defmacro", labels)

    def test_hover(self):
        """Test hover functionality."""
        server, _, _ = self.create_server()

        server._handle_initialize({"processId": 1234, "capabilities": {}})
        server._init_backend()

        # Define a function in the backend
        if server.backend:
            server.backend.eval("(defn my-test-fn [x] (+ x 1))")

        server._handle_did_open(
            {
                "textDocument": {
                    "uri": "file:///test.spork",
                    "languageId": "spork",
                    "version": 1,
                    "text": "(my-test-fn 5)",
                }
            }
        )

        result = server._handle_hover(
            {
                "textDocument": {"uri": "file:///test.spork"},
                "position": {"line": 0, "character": 5},
            }
        )

        # Should return hover info or None
        if result:
            self.assertIn("contents", result)

    def test_shutdown(self):
        """Test the shutdown request."""
        server, _, _ = self.create_server()

        server._handle_initialize({"processId": 1234, "capabilities": {}})

        self.assertFalse(server.shutdown_requested)

        server._handle_shutdown({})

        self.assertTrue(server.shutdown_requested)


class TestDiagnostics(unittest.TestCase):
    """Test diagnostic generation."""

    def test_parse_error_diagnostic(self):
        """Test that parse errors generate diagnostics."""
        from spork.lsp.protocol import JsonRpcProtocol, ProtocolReader, ProtocolWriter
        from spork.lsp.server import SporkLanguageServer

        input_buffer = io.BytesIO()
        output_buffer = io.BytesIO()

        protocol = JsonRpcProtocol(
            reader=ProtocolReader(input_buffer),
            writer=ProtocolWriter(output_buffer),
        )

        server = SporkLanguageServer(protocol=protocol)
        server._handle_initialize({"processId": 1234, "capabilities": {}})
        server._init_backend()

        # Open a document with a parse error (unclosed paren)
        server._handle_did_open(
            {
                "textDocument": {
                    "uri": "file:///test.spork",
                    "languageId": "spork",
                    "version": 1,
                    "text": "(defn broken [x",
                }
            }
        )

        # Check that diagnostics were published
        output_buffer.seek(0)
        # Read the response (if any was written)
        content = output_buffer.read()
        # Diagnostics should have been published
        # Note: The actual diagnostic content depends on the error format


if __name__ == "__main__":
    unittest.main()



"""
spork.runtime.core - Core runtime functions (Standard Library)

This module contains the core runtime functions that implement Spork's
standard library. These are the functions that compiled Spork code calls
at runtime.

Categories:
- Protocol system: protocol dispatch and registration
- Sequence operations: first, rest, seq, nth, conj, assoc, etc.
- Transient operations: transient, persistent!, conj!, etc.
- Lazy sequences: map, filter, take, drop, etc.
- Reducers: reduce, reductions, some, every, etc.
- Collection utilities: into, empty, count, etc.
- Math operations: inc, dec, +, -, *, /, etc.
- Bitwise operations: bit-and, bit-or, bit-xor, etc.
"""

from abc import ABC
from typing import Any, Iterator, Optional

from spork.runtime.pds import (
    EMPTY_MAP,
    EMPTY_SET,
    EMPTY_VECTOR,
    Cons,
    DoubleVector,
    IntVector,
    Map,
    Set,
    TransientDoubleVector,
    TransientIntVector,
    TransientMap,
    TransientSet,
    TransientVector,
    Vector,
    hash_map,
    vec,
)
from spork.runtime.types import _MISSING, Keyword

# =============================================================================
# Protocol System
# =============================================================================

# Global registries for protocols
_PROTOCOLS: dict[str, dict[str, Any]] = {
    # proto_name: {
    #   "abc": <ABC subclass>,
    #   "methods": ["first", "rest", "cons"],
    #   "structural": bool,
    #   "doc": str or None,
    # }
}

_PROTOCOL_IMPLS: dict[str, dict[type, dict[str, Any]]] = {
    # proto_name: {
    #   py_type: {
    #     "first": callable,
    #     "rest": callable,
    #   },
    # }
}


def runtime_register_protocol(
    name: str, doc: Optional[str], methods: list[str], structural: bool
) -> type:
    """
    Register a protocol and create its ABC.

    Args:
        name: Protocol name (e.g., "ISeq")
        doc: Optional docstring
        methods: List of method names
        structural: Whether to enable duck-typing fallback

    Returns:
        The ABC class for this protocol
    """
    # Create or fetch ABC
    abc_class = type(name, (ABC,), {"__doc__": doc or f"Protocol {name}"})
    _PROTOCOLS[name] = {
        "abc": abc_class,
        "methods": list(methods),
        "structural": structural,
        "doc": doc,
    }
    _PROTOCOL_IMPLS.setdefault(name, {})
    return abc_class


def get_protocol_abc(proto_name: str) -> Optional[type]:
    """Get the ABC class for a protocol by name."""
    proto = _PROTOCOLS.get(proto_name)
    if proto is None:
        return None
    return proto["abc"]


def register_protocol_impl(
    proto_name: str, py_type: type, methods_dict: dict[str, Any]
) -> None:
    """
    Register implementations of protocol methods for a type.

    Args:
        proto_name: Name of the protocol
        py_type: The Python type being extended
        methods_dict: Dict mapping method names to callables
    """
    if proto_name not in _PROTOCOLS:
        raise TypeError(f"Unknown protocol: {proto_name}")

    impls_for_proto = _PROTOCOL_IMPLS.setdefault(proto_name, {})
    # Convert keyword keys to string keys if needed
    normalized_dict = {}
    for k, v in methods_dict.items():
        if isinstance(k, str):
            key = k
        elif isinstance(k, Keyword):
            key = k.name  # Extract the name without the colon
        else:
            key = str(k)
        normalized_dict[key] = v
    impls_for_proto[py_type] = normalized_dict


def protocol_register_virtual_subclass(proto_name: str, py_type: type) -> None:
    """
    Register a type as a virtual subclass of a protocol's ABC.
    This enables isinstance checks.
    """
    if proto_name not in _PROTOCOLS:
        raise TypeError(f"Unknown protocol: {proto_name}")

    abc_class = _PROTOCOLS[proto_name]["abc"]
    abc_class.register(py_type)


def protocol_dispatch(proto_name: str, method_name: str, *args):
    """
    Dispatch a protocol method call.

    Args:
        proto_name: Name of the protocol
        method_name: Name of the method to call
        *args: Arguments to the method (first arg is dispatch target)

    Returns:
        Result of calling the implementation

    Raises:
        TypeError: If no implementation found
    """
    if not args:
        raise TypeError(f"{proto_name}.{method_name} called with no arguments")

    target = args[0]
    t = type(target)

    if proto_name not in _PROTOCOLS:
        raise TypeError(f"Unknown protocol: {proto_name}")

    proto = _PROTOCOLS[proto_name]
    impls = _PROTOCOL_IMPLS.get(proto_name, {})
    structural = proto["structural"]

    # 1. Exact type match
    entry = impls.get(t)
    if entry is not None:
        fn = entry.get(method_name)
        if fn is not None:
            return fn(*args)

    # 2. Walk MRO for supertype implementations
    for base in t.__mro__[1:]:
        entry = impls.get(base)
        if entry is not None:
            fn = entry.get(method_name)
            if fn is not None:
                return fn(*args)

    # 3. Optional structural duck-typing
    if structural:
        # Check if target has the method as an attribute
        # Convert method_name from Clojure style (e.g., "read-all") to Python style ("read_all")
        py_method_name = method_name.replace("-", "_")
        if hasattr(target, py_method_name):
            fn = getattr(target, py_method_name)
            # Call without target since it's a bound method
            return fn(*args[1:])

    # No implementation found
    raise TypeError(
        f"No implementation of {proto_name}.{method_name} for type {t.__name__}"
    )


def satisfies_protocol(proto_name: str, obj: Any) -> bool:
    """
    Check if an object satisfies a protocol.

    This checks:
    1. If the object's type has registered implementations
    2. If structural, whether the object has all required methods
    """
    if proto_name not in _PROTOCOLS:
        return False

    proto = _PROTOCOLS[proto_name]
    t = type(obj)
    impls = _PROTOCOL_IMPLS.get(proto_name, {})

    # Check exact type
    if t in impls:
        return True

    # Check MRO
    for base in t.__mro__[1:]:
        if base in impls:
            return True

    # Check structural (duck typing)
    if proto["structural"]:
        methods = proto["methods"]
        for method_name in methods:
            py_method_name = method_name.replace("-", "_")
            if not hasattr(obj, py_method_name):
                return False
        return True

    return False


# =============================================================================
# Sequence Operations (Core)
# =============================================================================


class LazySeq:
    """
    A lazy sequence wrapper that implements the seq abstraction without
    eagerly building Cons cells. Wraps an iterator and lazily materializes
    first/rest on demand.

    This is O(1) to create and only traverses the underlying iterator
    as elements are accessed.
    """

    __slots__ = ("_iterator", "_first", "_rest", "_realized")

    def __init__(self, iterator: Iterator):
        self._iterator = iterator
        self._first = None
        self._rest = None
        self._realized = False

    def _realize(self):
        """Realize the first element from the iterator."""
        if self._realized:
            return
        self._realized = True
        try:
            self._first = next(self._iterator)
        except StopIteration:
            # Empty iterator - mark as empty
            self._first = _MISSING
            self._rest = None

    @property
    def first(self):
        """Return the first element, realizing it if necessary."""
        self._realize()
        if self._first is _MISSING:
            return None
        return self._first

    @property
    def rest(self):
        """Return the rest as another LazySeq, or None if empty."""
        self._realize()
        if self._first is _MISSING:
            return None
        if self._rest is None:
            # Create a new LazySeq for the rest of the iterator
            self._rest = LazySeq(self._iterator)
            # Check if rest is empty
            self._rest._realize()
            if self._rest._first is _MISSING:
                self._rest = None
        return self._rest

    def __iter__(self):
        """Iterate through the lazy sequence."""
        current = self
        while current is not None:
            current._realize()
            if current._first is _MISSING:
                break
            yield current._first
            current = current.rest

    def __bool__(self):
        """LazySeq is truthy if it has at least one element."""
        self._realize()
        return self._first is not _MISSING

    def __repr__(self):
        # Collect elements for display (be careful with infinite seqs)
        items = []
        current = self
        limit = 10
        while current is not None and len(items) < limit:
            current._realize()
            if current._first is _MISSING:
                break
            items.append(repr(current._first))
            current = current.rest
        if current is not None and current._first is not _MISSING:
            items.append("...")
        return f"({' '.join(items)})"


def first(coll):
    """Return the first element of a collection."""
    if coll is None:
        return None
    if isinstance(coll, (Cons, LazySeq)):
        return coll.first
    if isinstance(coll, Vector):
        return coll.nth(0, None) if len(coll) > 0 else None
    # Fallback for Python sequences
    try:
        it = iter(coll)
        return next(it, None)
    except TypeError:
        return None


def rest(coll):
    """Return the rest of a collection as a sequence."""
    if coll is None:
        return None
    if isinstance(coll, (Cons, LazySeq)):
        r = coll.rest
        return r if r is not None and r is not type(None) else None
    if isinstance(coll, Vector):
        if len(coll) <= 1:
            return None
        # Return a Cons chain for rest of vector
        return seq(coll.nth(i) for i in range(1, len(coll)))
    # Fallback for Python sequences
    try:
        it = iter(coll)
        next(it, None)  # skip first
        return seq(it)
    except TypeError:
        return None


def seq(iterable):
    """Convert an iterable to a Cons sequence."""
    if iterable is None:
        return None
    if isinstance(iterable, Cons):
        return iterable
    if isinstance(iterable, LazySeq):
        # Realize LazySeq to Cons if needed, or just return it
        return iterable if iterable else None
    # For Map, return [key value] pairs
    if isinstance(iterable, Map):
        result = None
        for k, v in reversed(list(iterable.items())):
            result = Cons(vec(k, v), result)
        return result
    # Build in reverse then flip
    result = None
    items = list(iterable)
    for item in reversed(items):
        result = Cons(item, result)
    return result


def lazy_seq(iterable):
    """Convert an iterable to a lazy sequence.

    Returns a LazySeq that wraps the iterator, only materializing
    elements as they are accessed. This is O(1) to create.

    Unlike `seq`, this does not eagerly realize the entire sequence.
    """
    if iterable is None:
        return None
    if isinstance(iterable, LazySeq):
        return iterable
    if isinstance(iterable, Cons):
        # Wrap Cons iteration in LazySeq
        def cons_iter():
            curr = iterable
            while curr is not None:
                yield curr.first
                curr = curr.rest

        return LazySeq(cons_iter())
    # For Map, return [key value] pairs lazily
    if isinstance(iterable, Map):

        def map_pairs():
            for k, v in iterable.items():
                yield vec(k, v)

        return LazySeq(map_pairs())
    # Wrap in LazySeq - O(1) creation
    try:
        it = iter(iterable)
        result = LazySeq(it)
        # Check if empty
        if not result:
            return None
        return result
    except TypeError:
        return None


def nth(coll, index, default=_MISSING):
    """Get the nth element from any collection."""
    if coll is None:
        if default is _MISSING:
            raise IndexError(f"Index {index} out of range on nil")
        return default

    if isinstance(coll, Vector):
        return coll.nth(index, default) if default is not _MISSING else coll.nth(index)

    if isinstance(coll, (Cons, LazySeq)):
        curr = coll
        for _ in range(index):
            if curr is None:
                if default is _MISSING:
                    raise IndexError(f"Index {index} out of range")
                return default
            curr = curr.rest
        if curr is None:
            if default is _MISSING:
                raise IndexError(f"Index {index} out of range")
            return default
        return curr.first

    # Python sequences
    try:
        return coll[index]
    except (IndexError, KeyError):
        if default is _MISSING:
            raise
        return default


def conj(coll, val):
    """Add an element to a collection."""
    if coll is None:
        return Cons(val, None)
    if isinstance(coll, Vector):
        return coll.conj(val)
    if isinstance(coll, Cons):
        return Cons(val, coll)
    if isinstance(coll, Set):
        return coll.conj(val)
    if isinstance(coll, Map):
        if isinstance(val, Vector) and len(val) == 2:
            return coll.assoc(val.nth(0), val.nth(1))
        elif isinstance(val, (list, tuple)) and len(val) == 2:
            return coll.assoc(val[0], val[1])
        raise ValueError("conj on map requires a [key value] pair")
    # Fallback
    if hasattr(coll, "append"):
        new_coll = list(coll)
        new_coll.append(val)
        return new_coll
    raise TypeError(f"Don't know how to conj onto {type(coll)}")


def assoc(coll, key, val):
    """Associate a key with a value in a collection."""
    if coll is None:
        if isinstance(key, int):
            raise IndexError("Can't assoc on nil")
        return hash_map(key, val)
    if isinstance(coll, Vector):
        return coll.assoc(key, val)
    if isinstance(coll, Map):
        return coll.assoc(key, val)
    if isinstance(coll, dict):
        new_dict = dict(coll)
        new_dict[key] = val
        return new_dict
    raise TypeError(f"Don't know how to assoc onto {type(coll)}")


def dissoc(coll, key):
    """Remove a key from a map."""
    if coll is None:
        return None
    if isinstance(coll, Map):
        return coll.dissoc(key)
    if isinstance(coll, dict):
        new_dict = dict(coll)
        new_dict.pop(key, None)
        return new_dict
    raise TypeError(f"Don't know how to dissoc from {type(coll)}")


def disj(coll, val):
    """Remove an element from a set."""
    if coll is None:
        return None
    if isinstance(coll, Set):
        return coll.disj(val)
    if isinstance(coll, set):
        new_set = set(coll)
        new_set.discard(val)
        return new_set
    raise TypeError(f"Don't know how to disj from {type(coll)}")


def get(coll, key, default=None):
    """Get a value from a collection by key."""
    if coll is None:
        return default
    if isinstance(coll, Map):
        return coll.get(key, default)
    if isinstance(coll, Vector):
        if isinstance(key, int):
            return coll.nth(key, default)
        return default
    if isinstance(coll, dict):
        return coll.get(key, default)
    # Try indexing
    try:
        return coll[key]
    except (IndexError, KeyError, TypeError):
        return default


def count(coll):
    """Return the count of items in a collection."""
    if coll is None:
        return 0
    return len(coll)


def contains_q(coll, key):
    """Check if a collection contains a key/element.

    For maps: checks if key is present.
    For sets: checks if element is present.
    For vectors/lists: checks if index is valid (not value!).
    """
    if coll is None:
        return False
    if isinstance(coll, Set):
        return key in coll
    if isinstance(coll, Map):
        return key in coll
    if isinstance(coll, set):
        return key in coll
    if isinstance(coll, dict):
        return key in coll
    if isinstance(coll, (Vector, list, tuple)):
        # For indexed collections, contains? checks if index is valid
        if isinstance(key, int):
            return 0 <= key < len(coll)
        return False
    # Fallback: use 'in' operator
    return key in coll


def empty(coll):
    """Return an empty collection of the same type."""
    if coll is None:
        return None
    if isinstance(coll, Vector):
        return EMPTY_VECTOR
    if isinstance(coll, Map):
        return EMPTY_MAP
    if isinstance(coll, Set):
        return EMPTY_SET
    if isinstance(coll, Cons):
        return None
    if isinstance(coll, list):
        return []
    if isinstance(coll, dict):
        return {}
    if isinstance(coll, set):
        return set()
    return None


def into(to_coll, from_coll):
    """Add all items from from_coll into to_coll."""
    if to_coll is None:
        return seq(from_coll)
    if isinstance(to_coll, Vector):
        t = to_coll.transient()
        for item in from_coll:
            t.conj_mut(item)
        return t.persistent()
    if isinstance(to_coll, Map):
        t = to_coll.transient()
        for item in from_coll:
            if isinstance(item, Vector) and len(item) == 2:
                t.assoc_mut(item.nth(0), item.nth(1))
            elif isinstance(item, (list, tuple)) and len(item) == 2:
                t.assoc_mut(item[0], item[1])
            else:
                raise ValueError("into map requires [key value] pairs")
        return t.persistent()
    if isinstance(to_coll, Set):
        t = to_coll.transient()
        for item in from_coll:
            t.conj_mut(item)
        return t.persistent()
    if isinstance(to_coll, Cons):
        result = to_coll
        for item in from_coll:
            result = Cons(item, result)
        return result
    raise TypeError(f"Don't know how to into {type(to_coll)}")


# =============================================================================
# Transient Operations
# =============================================================================


def transient(coll):
    """Create a mutable (transient) version of a persistent collection.

    Transients allow efficient batch mutations. Use conj!, assoc!, dissoc!
    to mutate, then call persistent! to get back an immutable collection.
    """
    if isinstance(coll, Vector):
        return coll.transient()
    if isinstance(coll, DoubleVector):
        return coll.transient()
    if isinstance(coll, IntVector):
        return coll.transient()
    if isinstance(coll, Map):
        return coll.transient()
    if isinstance(coll, Set):
        return coll.transient()
    raise TypeError(f"Don't know how to create transient from {type(coll)}")


def persistent_bang(coll):
    """Convert a transient collection back to a persistent (immutable) one.

    After calling persistent!, the transient can no longer be used.
    """
    if isinstance(coll, TransientVector):
        return coll.persistent()
    if isinstance(coll, TransientDoubleVector):
        return coll.persistent()
    if isinstance(coll, TransientIntVector):
        return coll.persistent()
    if isinstance(coll, TransientMap):
        return coll.persistent()
    if isinstance(coll, TransientSet):
        return coll.persistent()
    raise TypeError(f"Don't know how to make persistent from {type(coll)}")


def conj_bang(coll, val):
    """Mutably add an element to a transient collection.

    Returns the same transient (mutated in place).
    """
    if isinstance(coll, TransientVector):
        return coll.conj_mut(val)
    if isinstance(coll, TransientDoubleVector):
        return coll.conj_mut(val)
    if isinstance(coll, TransientIntVector):
        return coll.conj_mut(val)
    if isinstance(coll, TransientMap):
        if isinstance(val, Vector) and len(val) == 2:
            return coll.assoc_mut(val.nth(0), val.nth(1))
        elif isinstance(val, (list, tuple)) and len(val) == 2:
            return coll.assoc_mut(val[0], val[1])
        raise ValueError("conj! on transient map requires a [key value] pair")
    if isinstance(coll, TransientSet):
        return coll.conj_mut(val)
    raise TypeError(f"Don't know how to conj! onto {type(coll)}")


def assoc_bang(coll, key, val):
    """Mutably associate a key with a value in a transient collection.

    Returns the same transient (mutated in place).
    """
    if isinstance(coll, TransientVector):
        return coll.assoc_mut(key, val)
    if isinstance(coll, TransientMap):
        return coll.assoc_mut(key, val)
    raise TypeError(f"Don't know how to assoc! onto {type(coll)}")


def dissoc_bang(coll, key):
    """Mutably remove a key from a transient map.

    Returns the same transient (mutated in place).
    """
    if isinstance(coll, TransientMap):
        return coll.dissoc_mut(key)
    raise TypeError(f"Don't know how to dissoc! from {type(coll)}")


def disj_bang(coll, val):
    """Mutably remove an element from a transient set.

    Returns the same transient (mutated in place).
    """
    if isinstance(coll, TransientSet):
        return coll.disj_mut(val)
    raise TypeError(f"Don't know how to disj! from {type(coll)}")


def pop_bang(coll):
    """Mutably remove the last element from a transient vector.

    Returns the same transient (mutated in place).
    """
    if isinstance(coll, TransientVector):
        return coll.pop_mut()
    raise TypeError(f"Don't know how to pop! from {type(coll)}")


# =============================================================================
# Lazy Sequence Functions (Generators)
# =============================================================================


def spork_map(f, *colls):
    """Lazily map a function over one or more collections (generator).

    With one collection: (map f coll) yields (f x) for each x in coll.
    With multiple collections: (map f c1 c2 ...) yields (f x1 x2 ...)
    where x1, x2, ... are from c1, c2, ... respectively.
    Stops when the shortest collection is exhausted.
    """
    if not colls:
        return
    if len(colls) == 1:
        for x in colls[0]:
            yield f(x)
    else:
        iterators = [iter(c) for c in colls]
        while True:
            try:
                args = [next(it) for it in iterators]
                yield f(*args)
            except StopIteration:
                return


def spork_filter(pred, coll):
    """Lazily filter a collection by a predicate (generator)."""
    if coll is None:
        return
    for x in coll:
        if pred(x):
            yield x


def take(n, coll):
    """Lazily take the first n elements from a collection (generator)."""
    if coll is None or n <= 0:
        return
    count = 0
    for x in coll:
        if count >= n:
            return
        yield x
        count += 1


def take_while(pred, coll):
    """Lazily take elements while predicate is true (generator)."""
    if coll is None:
        return
    for x in coll:
        if pred(x):
            yield x
        else:
            return


def drop(n, coll):
    """Lazily drop the first n elements from a collection (generator)."""
    if coll is None:
        return
    it = iter(coll)
    # Skip n items
    for _ in range(n):
        try:
            next(it)
        except StopIteration:
            return
    # Yield the rest
    yield from it


def drop_while(pred, coll):
    """Lazily drop elements while predicate is true, then yield the rest (generator)."""
    if coll is None:
        return
    dropping = True
    for x in coll:
        if dropping:
            if pred(x):
                continue
            else:
                dropping = False
        yield x


def concat(*colls):
    """Lazily concatenate multiple collections (generator)."""
    for coll in colls:
        if coll is not None:
            yield from coll


def spork_repeat(x, n=None):
    """Lazily repeat x, n times or infinitely if n is None (generator)."""
    if n is None:
        while True:
            yield x
    else:
        for _ in range(n):
            yield x


def cycle(coll):
    """Lazily cycle through a collection infinitely (generator).

    Note: Materializes the collection on first pass to enable cycling.
    """
    if coll is None:
        return
    items = list(coll)
    if not items:
        return
    while True:
        yield from items


def iterate(f, x):
    """Lazily generate x, f(x), f(f(x)), ... infinitely (generator)."""
    current = x
    while True:
        yield current
        current = f(current)


def spork_range(*args):
    """Lazily generate a range of numbers (generator).

    (range) -> 0, 1, 2, ... (infinite)
    (range end) -> 0, 1, ..., end-1
    (range start end) -> start, start+1, ..., end-1
    (range start end step) -> start, start+step, ..., up to end
    """
    if len(args) == 0:
        # Infinite range starting from 0
        n = 0
        while True:
            yield n
            n += 1
    elif len(args) == 1:
        yield from range(args[0])
    elif len(args) == 2:
        yield from range(args[0], args[1])
    elif len(args) == 3:
        yield from range(args[0], args[1], args[2])
    else:
        raise TypeError(f"range takes 0-3 arguments, got {len(args)}")


def interleave(*colls):
    """Lazily interleave multiple collections (generator).

    Stops when any collection is exhausted.
    """
    if not colls:
        return
    iterators = [iter(c) for c in colls]
    while True:
        try:
            for it in iterators:
                yield next(it)
        except StopIteration:
            return


def interpose(sep, coll):
    """Lazily interpose a separator between elements (generator)."""
    if coll is None:
        return
    first_elem = True
    for x in coll:
        if not first_elem:
            yield sep
        first_elem = False
        yield x


def partition(n, coll, step=None, pad=None):
    """Lazily partition a collection into chunks of n items (generator).

    step: how many items to advance (default: n)
    pad: collection to use to pad the last chunk if incomplete (default: None, drop incomplete)
    """
    if coll is None or n <= 0:
        return
    if step is None:
        step = n
    items = list(coll)
    i = 0
    while i < len(items):
        chunk = items[i : i + n]
        if len(chunk) == n:
            yield vec(*chunk)
        elif pad is not None:
            # Pad the last chunk
            pad_iter = iter(pad)
            while len(chunk) < n:
                try:
                    chunk.append(next(pad_iter))
                except StopIteration:
                    break
            if len(chunk) == n:
                yield vec(*chunk)
        # else: drop incomplete chunk
        i += step


def partition_all(n, coll, step=None):
    """Lazily partition a collection, including incomplete final chunk (generator)."""
    if coll is None or n <= 0:
        return
    if step is None:
        step = n
    items = list(coll)
    i = 0
    while i < len(items):
        chunk = items[i : i + n]
        yield vec(*chunk)
        i += step


def keep(f, coll):
    """Lazily keep non-nil results of f applied to coll (generator)."""
    if coll is None:
        return
    for x in coll:
        result = f(x)
        if result is not None:
            yield result


def keep_indexed(f, coll):
    """Lazily keep non-nil results of f(index, item) applied to coll (generator)."""
    if coll is None:
        return
    for i, x in enumerate(coll):
        result = f(i, x)
        if result is not None:
            yield result


def map_indexed(f, coll):
    """Lazily map f(index, item) over a collection (generator)."""
    if coll is None:
        return
    for i, x in enumerate(coll):
        yield f(i, x)


def dedupe(coll):
    """Lazily remove consecutive duplicates (generator)."""
    if coll is None:
        return
    prev = object()  # Unique sentinel
    for x in coll:
        if x != prev:
            yield x
            prev = x


def distinct(coll):
    """Lazily remove all duplicates, keeping first occurrence (generator)."""
    if coll is None:
        return
    seen = set()
    for x in coll:
        # Use id for unhashable types
        try:
            key = x
            if key not in seen:
                seen.add(key)
                yield x
        except TypeError:
            # Unhashable, use identity
            key = id(x)
            if key not in seen:
                seen.add(key)
                yield x


def flatten(coll):
    """Lazily flatten nested collections (generator).

    Flattens any iterable except strings.
    """
    if coll is None:
        return
    for x in coll:
        if hasattr(x, "__iter__") and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def mapcat(f, *colls):
    """Lazily map f over colls and concatenate results (generator).

    Equivalent to (concat (map f colls...)).
    """
    for result in spork_map(f, *colls):
        if result is not None:
            yield from result


# =============================================================================
# Sequence Predicates and Reducers
# =============================================================================


def some(pred, coll):
    """Return the first truthy value of pred(x) for x in coll, or None."""
    if coll is None:
        return None
    for x in coll:
        result = pred(x)
        if result:
            return result
    return None


def every(pred, coll):
    """Return True if pred(x) is truthy for all x in coll."""
    if coll is None:
        return True
    for x in coll:
        if not pred(x):
            return False
    return True


def not_every(pred, coll):
    """Return True if pred(x) is falsy for at least one x in coll."""
    return not every(pred, coll)


def not_any(pred, coll):
    """Return True if pred(x) is falsy for all x in coll."""
    if coll is None:
        return True
    for x in coll:
        if pred(x):
            return False
    return True


def reduce(f, *args):
    """Reduce a collection with a function.

    (reduce f coll) - reduces with no initial value
    (reduce f init coll) - reduces with initial value
    """
    if len(args) == 1:
        coll = args[0]
        it = iter(coll)
        try:
            acc = next(it)
        except StopIteration:
            return f()  # f with no args for empty collection
        for x in it:
            acc = f(acc, x)
        return acc
    elif len(args) == 2:
        init, coll = args
        acc = init
        for x in coll:
            acc = f(acc, x)
        return acc
    else:
        raise TypeError(f"reduce takes 2-3 arguments, got {len(args) + 1}")


def reductions(f, *args):
    """Lazily yield intermediate reduce values (generator).

    (reductions f coll) - with no initial value
    (reductions f init coll) - with initial value
    """
    if len(args) == 1:
        coll = args[0]
        it = iter(coll)
        try:
            acc = next(it)
        except StopIteration:
            return
        yield acc
        for x in it:
            acc = f(acc, x)
            yield acc
    elif len(args) == 2:
        init, coll = args
        acc = init
        yield acc
        for x in coll:
            acc = f(acc, x)
            yield acc
    else:
        raise TypeError(f"reductions takes 2-3 arguments, got {len(args) + 1}")


# =============================================================================
# Collection Utilities
# =============================================================================


def zipmap(keys, vals):
    """Return a Map with keys mapped to corresponding vals."""
    result = EMPTY_MAP
    if result is None:
        # pds not initialized yet, return dict
        return dict(zip(keys, vals))
    t = result.transient()
    for k, v in zip(keys, vals):
        t.assoc_mut(k, v)
    return t.persistent()


def group_by(f, coll):
    """Return a Map of elements grouped by the result of f."""
    result = {}
    for x in coll:
        key = f(x)
        if key not in result:
            result[key] = []
        result[key].append(x)
    # Convert to Map with Vector values
    if EMPTY_MAP is not None:
        t = EMPTY_MAP.transient()
        for k, v in result.items():
            t.assoc_mut(k, vec(*v))
        return t.persistent()
    return result


def frequencies(coll):
    """Return a Map of elements to their counts."""
    result = {}
    for x in coll:
        result[x] = result.get(x, 0) + 1
    if EMPTY_MAP is not None:
        t = EMPTY_MAP.transient()
        for k, v in result.items():
            t.assoc_mut(k, v)
        return t.persistent()
    return result


def reverse(coll):
    """Return a lazy reversed sequence (realizes collection first)."""
    if coll is None:
        return
    items = list(coll)
    for i in range(len(items) - 1, -1, -1):
        yield items[i]


def sort(coll, key=None, reverse_order=False):
    """Return a sorted sequence (realizes collection)."""
    if coll is None:
        return EMPTY_VECTOR if EMPTY_VECTOR is not None else []
    items = sorted(coll, key=key, reverse=reverse_order)
    return vec(*items) if vec is not None else items


def sort_by(keyfn, coll):
    """Return a sequence sorted by keyfn (realizes collection)."""
    return sort(coll, key=keyfn)


def split_at(n, coll):
    """Return a tuple of (take n coll, drop n coll)."""
    if coll is None:
        return (vec(), vec()) if vec else ([], [])
    items = list(coll)
    return (vec(*items[:n]), vec(*items[n:])) if vec else (items[:n], items[n:])


def split_with(pred, coll):
    """Return a tuple of (take-while pred coll, drop-while pred coll)."""
    if coll is None:
        return (vec(), vec()) if vec else ([], [])
    items = list(coll)
    i = 0
    for x in items:
        if pred(x):
            i += 1
        else:
            break
    return (vec(*items[:i]), vec(*items[i:])) if vec else (items[:i], items[i:])


def doall(coll):
    """Force realization of a lazy sequence, return as Vector."""
    if coll is None:
        return EMPTY_VECTOR if EMPTY_VECTOR is not None else []
    return vec(*coll) if vec is not None else list(coll)


def dorun(coll):
    """Force realization of a lazy sequence for side effects only, return None."""
    if coll is None:
        return None
    for _ in coll:
        pass
    return None


def realized_q(coll):
    """Return True if the collection is fully realized (not a generator)."""
    import types

    return not isinstance(coll, types.GeneratorType)


# =============================================================================
# Math Operations
# =============================================================================


def inc(x):
    """Increment a number by 1."""
    return x + 1


def dec(x):
    """Decrement a number by 1."""
    return x - 1


def even_q(x):
    """Return True if x is even."""
    return x % 2 == 0


def odd_q(x):
    """Return True if x is odd."""
    return x % 2 == 1


def pos_q(x):
    """Return True if x is positive."""
    return x > 0


def neg_q(x):
    """Return True if x is negative."""
    return x < 0


def zero_q(x):
    """Return True if x is zero."""
    return x == 0


def add(*args):
    """Add numbers together."""
    if len(args) == 0:
        return 0
    result = args[0]
    for x in args[1:]:
        result = result + x
    return result


def sub(*args):
    """Subtract numbers."""
    if len(args) == 0:
        return 0
    if len(args) == 1:
        return -args[0]
    result = args[0]
    for x in args[1:]:
        result = result - x
    return result


def mul(*args):
    """Multiply numbers together."""
    if len(args) == 0:
        return 1
    result = args[0]
    for x in args[1:]:
        result = result * x
    return result


def div(*args):
    """Divide numbers."""
    if len(args) == 0:
        raise TypeError("div requires at least 1 argument")
    if len(args) == 1:
        return 1 / args[0]
    result = args[0]
    for x in args[1:]:
        result = result / x
    return result


def mod(a, b):
    """Modulo operation."""
    return a % b


def quot(a, b):
    """Integer division (quotient)."""
    return a // b


def spork_max(*args):
    """Return the maximum value."""
    if len(args) == 1 and hasattr(args[0], "__iter__"):
        return max(args[0])
    return max(args)


def spork_min(*args):
    """Return the minimum value."""
    if len(args) == 1 and hasattr(args[0], "__iter__"):
        return min(args[0])
    return min(args)


def spork_abs(x):
    """Return the absolute value."""
    return abs(x)


# =============================================================================
# Bitwise Operations
# =============================================================================


def bit_or(*args):
    """Bitwise OR. Also works as set union."""
    if len(args) == 0:
        return 0
    result = args[0]
    for x in args[1:]:
        result = result | x
    return result


def bit_and(*args):
    """Bitwise AND. Also works as set intersection."""
    if len(args) < 2:
        raise TypeError("bit-and requires at least 2 arguments")
    result = args[0]
    for x in args[1:]:
        result = result & x
    return result


def bit_and_not(*args):
    """Bitwise AND NOT. Also works as set difference."""
    if len(args) < 2:
        raise TypeError("bit-and-not requires at least 2 arguments")
    result = args[0]
    for x in args[1:]:
        result = result & (~x)
    return result


def bit_xor(*args):
    """Bitwise XOR. Also works as set symmetric difference."""
    if len(args) == 0:
        return 0
    result = args[0]
    for x in args[1:]:
        result = result ^ x
    return result


def bit_not(x):
    """Bitwise NOT (complement)."""
    return ~x


def bit_shift_left(x, n):
    """Bitwise left shift."""
    return x << n


def bit_shift_right(x, n):
    """Bitwise right shift."""
    return x >> n


# =============================================================================
# Exports
# =============================================================================

__all__ = [
    # Protocol system
    "_PROTOCOLS",
    "_PROTOCOL_IMPLS",
    "runtime_register_protocol",
    "get_protocol_abc",
    "register_protocol_impl",
    "protocol_register_virtual_subclass",
    "protocol_dispatch",
    "satisfies_protocol",
    # Sequence operations (core)
    "LazySeq",
    "lazy_seq",
    "first",
    "rest",
    "seq",
    "nth",
    "conj",
    "assoc",
    "dissoc",
    "disj",
    "get",
    "count",
    "contains_q",
    "empty",
    "into",
    # Transient operations
    "transient",
    "persistent_bang",
    "conj_bang",
    "assoc_bang",
    "dissoc_bang",
    "disj_bang",
    "pop_bang",
    # Lazy sequences
    "spork_map",
    "spork_filter",
    "take",
    "take_while",
    "drop",
    "drop_while",
    "concat",
    "spork_repeat",
    "cycle",
    "iterate",
    "spork_range",
    "interleave",
    "interpose",
    "partition",
    "partition_all",
    "keep",
    "keep_indexed",
    "map_indexed",
    "dedupe",
    "distinct",
    "flatten",
    "mapcat",
    # Predicates and reducers
    "some",
    "every",
    "not_every",
    "not_any",
    "reduce",
    "reductions",
    # Collection utilities
    "zipmap",
    "group_by",
    "frequencies",
    "reverse",
    "sort",
    "sort_by",
    "split_at",
    "split_with",
    "doall",
    "dorun",
    "realized_q",
    # Math operations
    "inc",
    "dec",
    "even_q",
    "odd_q",
    "pos_q",
    "neg_q",
    "zero_q",
    "add",
    "sub",
    "mul",
    "div",
    "mod",
    "quot",
    "spork_max",
    "spork_min",
    "spork_abs",
    # Bitwise operations
    "bit_or",
    "bit_and",
    "bit_and_not",
    "bit_xor",
    "bit_not",
    "bit_shift_left",
    "bit_shift_right",
]



"""
spork.runtime - The Spork Runtime Library

This package contains everything needed to run compiled Spork programs.
It is designed to be lightweight and independent of the compiler.

Submodules:
- types: Core type definitions (Symbol, Keyword, VectorLiteral, etc.)
- pds: C extension persistent data structures
- core: Standard library functions (first, rest, map, filter, etc.)
- utils: Runtime utilities (spork_try, setup_runtime_env, etc.)
- ns: Namespace management (loading, registering, finding namespaces)

The runtime is what gets installed into user project environments.
The compiler (spork.compiler) depends on runtime.types for AST nodes,
but the runtime has no dependencies on the compiler.
"""

from typing import Any, TypeVar, Union

# Re-export core functions
from spork.runtime.core import (
    _PROTOCOL_IMPLS,
    _PROTOCOLS,
    LazySeq,
    add,
    assoc,
    assoc_bang,
    bit_and,
    bit_and_not,
    bit_not,
    bit_or,
    bit_shift_left,
    bit_shift_right,
    bit_xor,
    concat,
    conj,
    conj_bang,
    contains_q,
    count,
    cycle,
    dec,
    dedupe,
    disj,
    disj_bang,
    dissoc,
    dissoc_bang,
    distinct,
    div,
    doall,
    dorun,
    drop,
    drop_while,
    empty,
    even_q,
    every,
    first,
    flatten,
    frequencies,
    get,
    get_protocol_abc,
    group_by,
    inc,
    interleave,
    interpose,
    into,
    iterate,
    keep,
    keep_indexed,
    lazy_seq,
    map_indexed,
    mapcat,
    mod,
    mul,
    neg_q,
    not_any,
    not_every,
    nth,
    odd_q,
    partition,
    partition_all,
    persistent_bang,
    pop_bang,
    pos_q,
    protocol_dispatch,
    protocol_register_virtual_subclass,
    quot,
    realized_q,
    reduce,
    reductions,
    register_protocol_impl,
    rest,
    reverse,
    runtime_register_protocol,
    satisfies_protocol,
    seq,
    some,
    sort,
    sort_by,
    split_at,
    split_with,
    spork_abs,
    spork_filter,
    spork_map,
    spork_max,
    spork_min,
    spork_range,
    spork_repeat,
    sub,
    take,
    take_while,
    transient,
    zero_q,
    zipmap,
)

# Re-export JSON utilities
from spork.runtime.json import (
    SporkJSONEncoder,
)
from spork.runtime.json import (
    dump as json_dump,
)
from spork.runtime.json import (
    dumps as json_dumps,
)
from spork.runtime.json import (
    load as json_load,
)
from spork.runtime.json import (
    load_spork as json_load_spork,
)
from spork.runtime.json import (
    loads as json_loads,
)
from spork.runtime.json import (
    loads_spork as json_loads_spork,
)

# Re-export PDS (persistent data structures)
from spork.runtime.pds import (  # pyright: ignore[reportMissingModuleSource]
    EMPTY_DOUBLE_VECTOR,
    EMPTY_LONG_VECTOR,
    EMPTY_MAP,
    EMPTY_SET,
    EMPTY_VECTOR,
    Cons,
    DoubleVector,
    IntVector,
    Map,
    Set,
    TransientDoubleVector,
    TransientIntVector,
    TransientMap,
    TransientSet,
    TransientVector,
    Vector,
    cons,
    hash_map,
    hash_set,
    vec,
    vec_f64,
    vec_i64,
)

# Re-export types
from spork.runtime.types import (
    _MISSING,
    Decorated,
    Keyword,
    KwargsLiteral,
    MapLiteral,
    MatchError,
    SetLiteral,
    Symbol,
    VectorLiteral,
)

# Re-export utils
from spork.runtime.utils import (
    __spork_ns_env__,
    __spork_ns_get__,
    __spork_refer_all__,
    __spork_require__,
    setup_runtime_env,
    spork_raise,
    spork_setattr,
    spork_try,
)

# Type variables for generic types
T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
T_co = TypeVar("T_co", covariant=True)
V_co = TypeVar("V_co", covariant=True)

# Type aliases for common patterns
List = Vector
Vec = Vector
Dict = Map
ConsCell = Cons


def is_typed_collection(obj: Any) -> bool:
    """Check if an object is one of Spork's typed persistent collections."""
    return isinstance(obj, (Vector, Map, Cons))


def get_collection_type(obj: Any) -> Union[type, None]:
    """Get the collection type of a Spork persistent data structure."""
    if isinstance(obj, Vector):
        return Vector
    elif isinstance(obj, Map):
        return Map
    elif isinstance(obj, Cons):
        return Cons
    return None


__all__ = [
    # Types
    "Symbol",
    "Keyword",
    "VectorLiteral",
    "MapLiteral",
    "SetLiteral",
    "KwargsLiteral",
    "Decorated",
    "MatchError",
    "_MISSING",
    # Persistent data structures
    "Vector",
    "Map",
    "Set",
    "Cons",
    "DoubleVector",
    "IntVector",
    "TransientVector",
    "TransientDoubleVector",
    "TransientIntVector",
    "TransientMap",
    "TransientSet",
    "EMPTY_VECTOR",
    "EMPTY_DOUBLE_VECTOR",
    "EMPTY_LONG_VECTOR",
    "EMPTY_MAP",
    "EMPTY_SET",
    "vec",
    "vec_f64",
    "vec_i64",
    "hash_map",
    "hash_set",
    "cons",
    # JSON
    "SporkJSONEncoder",
    "json_dump",
    "json_dumps",
    "json_load",
    "json_loads",
    "json_load_spork",
    "json_loads_spork",
    # Protocol system
    "_PROTOCOLS",
    "_PROTOCOL_IMPLS",
    "runtime_register_protocol",
    "get_protocol_abc",
    "register_protocol_impl",
    "protocol_register_virtual_subclass",
    "protocol_dispatch",
    "satisfies_protocol",
    # Sequence operations
    "first",
    "rest",
    "seq",
    "nth",
    "conj",
    "assoc",
    "dissoc",
    "disj",
    "get",
    "count",
    "contains_q",
    "empty",
    "into",
    # Transients
    "transient",
    "persistent_bang",
    "conj_bang",
    "assoc_bang",
    "dissoc_bang",
    "disj_bang",
    "pop_bang",
    # Lazy sequences
    "spork_map",
    "spork_filter",
    "take",
    "take_while",
    "drop",
    "drop_while",
    "concat",
    "spork_repeat",
    "cycle",
    "iterate",
    "spork_range",
    "interleave",
    "interpose",
    "partition",
    "partition_all",
    "keep",
    "keep_indexed",
    "map_indexed",
    "dedupe",
    "distinct",
    "flatten",
    "mapcat",
    # Predicates and reducers
    "some",
    "every",
    "not_every",
    "not_any",
    "reduce",
    "reductions",
    # Collection utilities
    "zipmap",
    "group_by",
    "frequencies",
    "reverse",
    "sort",
    "sort_by",
    "split_at",
    "split_with",
    "doall",
    "dorun",
    "realized_q",
    # Math
    "inc",
    "dec",
    "even_q",
    "odd_q",
    "pos_q",
    "neg_q",
    "zero_q",
    "add",
    "sub",
    "mul",
    "div",
    "mod",
    "quot",
    "spork_max",
    "spork_min",
    "spork_abs",
    # Bitwise
    "bit_or",
    "bit_and",
    "bit_and_not",
    "bit_xor",
    "bit_not",
    "bit_shift_left",
    "bit_shift_right",
    # Utils
    "spork_try",
    "spork_raise",
    "spork_setattr",
    "setup_runtime_env",
    "__spork_require__",
    "__spork_ns_env__",
    "__spork_ns_get__",
    "__spork_refer_all__",
    # Type variables
    "T",
    "T_co",
    "K",
    "V",
    "V_co",
    # Type aliases
    "List",
    "Vec",
    "Dict",
    "ConsCell",
    # Utilities
    "is_typed_collection",
    "get_collection_type",
]



"""
spork.lsp.server - Spork Language Server Implementation

This module provides the main LSP server for Spork. It leverages the
ReplBackend for evaluation and symbol information, and integrates with
the project system for workspace awareness.

Features:
- Code completion
- Hover documentation
- Go to definition
- Diagnostics (parse errors, compile errors)
- Document synchronization

Usage:
    The server is started via `spork lsp` and communicates over stdio.
"""

import os
import sys
import traceback
from dataclasses import dataclass, field
from typing import Any, Optional

from spork.lsp.protocol import (
    CompletionItemKind,
    DiagnosticSeverity,
    ErrorCode,
    JsonRpcError,
    JsonRpcProtocol,
    ProtocolReader,
    ProtocolWriter,
    TextDocumentSyncKind,
    make_completion_item,
    make_diagnostic,
    make_hover,
    make_location,
    make_range,
    path_to_uri,
    uri_to_path,
)


@dataclass
class TextDocument:
    """Represents an open text document."""

    uri: str
    language_id: str
    version: int
    content: str

    @property
    def path(self) -> str:
        """Get the file system path for this document."""
        return uri_to_path(self.uri)

    def get_line(self, line: int) -> str:
        """Get a specific line (0-based)."""
        lines = self.content.split("\n")
        if 0 <= line < len(lines):
            return lines[line]
        return ""

    def get_word_at_position(self, line: int, character: int) -> Optional[str]:
        """
        Get the word at the given position.

        Returns the symbol/word under or before the cursor.
        """
        line_text = self.get_line(line)
        if not line_text or character > len(line_text):
            return None

        # Find word boundaries
        # Spork symbols can contain: letters, digits, -, _, /, ., ?, !, +, *, <, >, =, &, #, ^
        # Based on LANG.md: identifiers like my-variable, valid?, math/sin are valid
        def is_symbol_char(c: str) -> bool:
            return c.isalnum() or c in "-_/.?!+*<>=:&^#"

        # Find start of word
        start = character
        while start > 0 and is_symbol_char(line_text[start - 1]):
            start -= 1

        # Find end of word
        end = character
        while end < len(line_text) and is_symbol_char(line_text[end]):
            end += 1

        if start == end:
            return None

        return line_text[start:end]

    def offset_to_position(self, offset: int) -> tuple[int, int]:
        """Convert a character offset to (line, character)."""
        line = 0
        col = 0
        for i, c in enumerate(self.content):
            if i == offset:
                break
            if c == "\n":
                line += 1
                col = 0
            else:
                col += 1
        return line, col


@dataclass
class SporkLanguageServer:
    """
    Spork Language Server Protocol implementation.

    This server provides IDE features for Spork source files.
    """

    # Protocol handler
    protocol: JsonRpcProtocol = field(default_factory=JsonRpcProtocol)

    # Open documents: uri -> TextDocument
    documents: dict[str, TextDocument] = field(default_factory=dict)

    # REPL backend for evaluation and symbol info
    backend: Any = None

    # Project configuration (if in a project)
    project_config: Any = None
    project_manager: Any = None

    # Server state
    initialized: bool = False
    shutdown_requested: bool = False

    # Workspace root
    root_uri: Optional[str] = None
    root_path: Optional[str] = None

    # Logging
    log_file: Any = None

    def __post_init__(self):
        """Set up the server after initialization."""
        self._register_handlers()

    def _log(self, message: str) -> None:
        """Log a message for debugging."""
        if self.log_file:
            self.log_file.write(f"{message}\n")
            self.log_file.flush()
        # Also write to stderr for debugging (LSP clients typically capture this)
        print(f"[spork-lsp] {message}", file=sys.stderr)
        sys.stderr.flush()

    def _register_handlers(self) -> None:
        """Register all LSP method handlers."""
        # Lifecycle
        self.protocol.register_request_handler("initialize", self._handle_initialize)
        self.protocol.register_notification_handler(
            "initialized", self._handle_initialized
        )
        self.protocol.register_request_handler("shutdown", self._handle_shutdown)
        self.protocol.register_notification_handler("exit", self._handle_exit)

        # Text document synchronization
        self.protocol.register_notification_handler(
            "textDocument/didOpen", self._handle_did_open
        )
        self.protocol.register_notification_handler(
            "textDocument/didChange", self._handle_did_change
        )
        self.protocol.register_notification_handler(
            "textDocument/didClose", self._handle_did_close
        )
        self.protocol.register_notification_handler(
            "textDocument/didSave", self._handle_did_save
        )

        # Language features
        self.protocol.register_request_handler(
            "textDocument/completion", self._handle_completion
        )
        self.protocol.register_request_handler("textDocument/hover", self._handle_hover)
        self.protocol.register_request_handler(
            "textDocument/definition", self._handle_definition
        )
        self.protocol.register_request_handler(
            "textDocument/references", self._handle_references
        )
        self.protocol.register_request_handler(
            "textDocument/documentSymbol", self._handle_document_symbol
        )

    # =========================================================================
    # Lifecycle Methods
    # =========================================================================

    def _handle_initialize(self, params: dict[str, Any]) -> dict[str, Any]:
        """Handle the initialize request."""
        self._log("Received initialize request")

        # Get workspace root
        self.root_uri = params.get("rootUri")
        if self.root_uri:
            self.root_path = uri_to_path(self.root_uri)
        else:
            # Fall back to rootPath (deprecated but still used)
            self.root_path = params.get("rootPath")
            if self.root_path:
                self.root_uri = path_to_uri(self.root_path)

        self._log(f"Workspace root: {self.root_path}")

        # Try to load project configuration
        self._load_project()

        # Initialize the REPL backend
        self._init_backend()

        # Return server capabilities
        return {
            "capabilities": {
                "textDocumentSync": {
                    "openClose": True,
                    "change": TextDocumentSyncKind.FULL,
                    "save": {"includeText": True},
                },
                "completionProvider": {
                    "triggerCharacters": [".", "/", ":"],
                    "resolveProvider": False,
                },
                "hoverProvider": True,
                "definitionProvider": True,
                "referencesProvider": False,  # Not yet implemented
                "documentSymbolProvider": True,
            },
            "serverInfo": {
                "name": "spork-lsp",
                "version": "0.1.0",
            },
        }

    def _handle_initialized(self, params: dict[str, Any]) -> None:
        """Handle the initialized notification."""
        self._log("Server initialized")
        self.initialized = True

        # If we have a project, we could scan for files here
        if self.project_config:
            self._log(f"Project: {self.project_config.name}")

    def _handle_shutdown(self, params: dict[str, Any]) -> None:
        """Handle the shutdown request."""
        self._log("Shutdown requested")
        self.shutdown_requested = True
        return None

    def _handle_exit(self, params: dict[str, Any]) -> None:
        """Handle the exit notification."""
        self._log("Exit notification received")
        exit_code = 0 if self.shutdown_requested else 1
        sys.exit(exit_code)

    # =========================================================================
    # Document Synchronization
    # =========================================================================

    def _handle_did_open(self, params: dict[str, Any]) -> None:
        """Handle textDocument/didOpen notification."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")
        language_id = text_document.get("languageId", "spork")
        version = text_document.get("version", 0)
        content = text_document.get("text", "")

        self._log(f"Document opened: {uri}")

        doc = TextDocument(
            uri=uri,
            language_id=language_id,
            version=version,
            content=content,
        )
        self.documents[uri] = doc

        # Validate the document
        self._validate_document(doc)

    def _handle_did_change(self, params: dict[str, Any]) -> None:
        """Handle textDocument/didChange notification."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")
        version = text_document.get("version", 0)
        content_changes = params.get("contentChanges", [])

        if uri not in self.documents:
            self._log(f"Warning: didChange for unknown document: {uri}")
            return

        doc = self.documents[uri]

        # We're using full sync, so take the last content change
        if content_changes:
            doc.content = content_changes[-1].get("text", doc.content)
        doc.version = version

        # Re-validate
        self._validate_document(doc)

    def _handle_did_close(self, params: dict[str, Any]) -> None:
        """Handle textDocument/didClose notification."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")

        self._log(f"Document closed: {uri}")

        if uri in self.documents:
            del self.documents[uri]

        # Clear diagnostics for closed document
        self._publish_diagnostics(uri, [])

    def _handle_did_save(self, params: dict[str, Any]) -> None:
        """Handle textDocument/didSave notification."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")
        text = params.get("text")

        self._log(f"Document saved: {uri}")

        if uri in self.documents and text is not None:
            self.documents[uri].content = text
            self._validate_document(self.documents[uri])

    # =========================================================================
    # Language Features
    # =========================================================================

    def _handle_completion(self, params: dict[str, Any]) -> dict[str, Any]:
        """Handle textDocument/completion request."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")
        position = params.get("position", {})
        line = position.get("line", 0)
        character = position.get("character", 0)

        if uri not in self.documents:
            return {"isIncomplete": False, "items": []}

        doc = self.documents[uri]
        prefix = self._get_completion_prefix(doc, line, character)

        self._log(f"Completion requested at {line}:{character}, prefix: '{prefix}'")

        items = []

        if self.backend:
            try:
                # Get completions from backend
                completions = self.backend.get_completions(prefix)

                for name in completions:
                    # Determine the kind based on symbol info
                    info = self.backend.get_symbol_info(name)
                    sym_type = info.get("type", "var")

                    if sym_type == "function":
                        kind = CompletionItemKind.FUNCTION
                    elif sym_type == "macro":
                        kind = CompletionItemKind.KEYWORD
                    elif sym_type == "class":
                        kind = CompletionItemKind.CLASS
                    elif sym_type == "protocol":
                        kind = CompletionItemKind.INTERFACE
                    elif sym_type == "protocol-fn":
                        kind = CompletionItemKind.METHOD
                    else:
                        kind = CompletionItemKind.VARIABLE

                    doc_string = info.get("doc")
                    detail = info.get("ns")

                    items.append(
                        make_completion_item(
                            label=name,
                            kind=kind,
                            detail=detail,
                            documentation=doc_string,
                        )
                    )
            except Exception as e:
                self._log(f"Completion error: {e}")

        # Add special forms and keywords if no prefix or matching
        # These are the actual special forms from spork/compiler/codegen.py
        special_forms = [
            # Definition forms (from compile_toplevel and compile_stmt)
            "def",
            "defn",
            "defmacro",
            "defclass",
            "fn",
            "let",
            "set!",
            # Control flow (from compile_stmt and compile_expr)
            "if",
            "do",
            "loop",
            "recur",
            "for",
            "while",
            "async-for",
            # Exception handling
            "try",
            "catch",
            "finally",
            "throw",
            "return",
            # Quoting (from compile_expr)
            "quote",
            "quasiquote",
            # Namespace/modules (from compile_toplevel)
            "ns",
            "import",
            # Async/generators (from compile_stmt)
            "await",
            "yield",
            "yield-from",
            # Pattern matching (from compile_expr)
            "match",
            # Python interop
            "with",
            "apply",
            "call",
            # Attribute access
            ".",
        ]

        # Macros from spork/std/prelude.spork
        prelude_macros = [
            # Control flow macros
            "when",
            "unless",
            "cond",
            # Threading macros
            "->",
            "->>",
            # Utility macros
            "comment",
            "fmt",
            "assert",
            # Sequence macros
            "mapv",
            "filterv",
            "doseq",
            "for-all",
            # Function composition
            "comp",
            "partial",
            "identity",
            "constantly",
            "complement",
            # Predicates
            "nil?",
            "some?",
            "string?",
            "number?",
            "int?",
            "float?",
            "bool?",
            "fn?",
            "symbol?",
            "keyword?",
            "vector?",
            "map?",
            "list?",
            "seq?",
            "coll?",
            "dict?",
            "empty?",
            "not-empty",
            "even?",
            "odd?",
            "pos?",
            "neg?",
            "zero?",
            # Collection accessors
            "second",
            "ffirst",
            "last",
            "butlast",
            # Protocol macros
            "defprotocol",
            "extend-type",
            "extend-protocol",
        ]

        for kw in special_forms:
            if kw.startswith(prefix):
                items.append(
                    make_completion_item(
                        label=kw,
                        kind=CompletionItemKind.KEYWORD,
                        detail="special form",
                    )
                )

        for macro in prelude_macros:
            if macro.startswith(prefix):
                items.append(
                    make_completion_item(
                        label=macro,
                        kind=CompletionItemKind.KEYWORD,
                        detail="macro",
                    )
                )

        return {"isIncomplete": False, "items": items}

    def _handle_hover(self, params: dict[str, Any]) -> Optional[dict[str, Any]]:
        """Handle textDocument/hover request."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")
        position = params.get("position", {})
        line = position.get("line", 0)
        character = position.get("character", 0)

        if uri not in self.documents:
            return None

        doc = self.documents[uri]
        word = doc.get_word_at_position(line, character)

        if not word:
            return None

        self._log(f"Hover requested for: {word}")

        if not self.backend:
            return None

        try:
            # Get symbol info from backend
            info = self.backend.get_symbol_info(word)

            if info.get("status") == "not-found":
                return None

            # Build hover content in Markdown
            parts = []

            sym_type = info.get("type", "unknown")
            ns = info.get("ns", "")

            # Type and name (Spork uses / as namespace separator like Clojure)
            if ns:
                parts.append(f"**{ns}/{word}** ({sym_type})")
            else:
                parts.append(f"**{word}** ({sym_type})")

            # Arglists for functions - show in Spork style with brackets
            if "arglists" in info:
                for arglist in info["arglists"]:
                    # Spork uses [arg1 arg2] for parameter vectors
                    args = " ".join(arglist)
                    parts.append(f"\n```spork\n({word} [{args}])\n```")

            # Documentation
            if info.get("doc"):
                parts.append(f"\n{info['doc']}")

            # Protocol info
            if sym_type == "protocol-fn" and info.get("protocol"):
                parts.append(f"\nProtocol: {info['protocol']}")

            if info.get("impls"):
                parts.append(f"\nImplemented by: {', '.join(info['impls'])}")

            content = "\n".join(parts)
            return make_hover(content)

        except Exception as e:
            self._log(f"Hover error: {e}")
            return None

    def _handle_definition(self, params: dict[str, Any]) -> Optional[dict[str, Any]]:
        """Handle textDocument/definition request."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")
        position = params.get("position", {})
        line = position.get("line", 0)
        character = position.get("character", 0)

        if uri not in self.documents:
            return None

        doc = self.documents[uri]
        word = doc.get_word_at_position(line, character)

        if not word:
            return None

        self._log(f"Definition requested for: {word}")

        if not self.backend:
            return None

        try:
            # Get source location from backend
            location = self.backend.get_source_location(word)

            if not location:
                return None

            file_path = location.get("file")
            source_line = location.get("line", 1)
            source_col = location.get("col", 0)

            if not file_path or not os.path.isfile(file_path):
                return None

            # LSP uses 0-based line numbers, but inspect uses 1-based
            return make_location(
                uri=path_to_uri(file_path),
                range_=make_range(
                    source_line - 1, source_col, source_line - 1, source_col
                ),
            )

        except Exception as e:
            self._log(f"Definition error: {e}")
            return None

    def _handle_references(self, params: dict[str, Any]) -> list[dict[str, Any]]:
        """Handle textDocument/references request."""
        # Not yet implemented - would require indexing the project
        return []

    def _handle_document_symbol(self, params: dict[str, Any]) -> list[dict[str, Any]]:
        """Handle textDocument/documentSymbol request."""
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri", "")

        if uri not in self.documents:
            return []

        doc = self.documents[uri]
        symbols = []

        try:
            # Parse the document to find definitions
            from spork.compiler.reader import read_str

            forms = read_str(doc.content)

            for form in forms:
                symbol = self._extract_definition_symbol(form)
                if symbol:
                    symbols.append(symbol)

        except Exception as e:
            self._log(f"Document symbol error: {e}")

        return symbols

    def _extract_definition_symbol(self, form: Any) -> Optional[dict[str, Any]]:
        """Extract a symbol definition from a form if it's a def/defn/etc."""
        from spork.lsp.protocol import SymbolKind
        from spork.runtime.types import Symbol

        if not isinstance(form, list) or len(form) < 2:
            return None

        head = form[0]
        if not isinstance(head, Symbol):
            return None

        head_name = head.name

        # Map form type to symbol kind
        # Based on actual forms in spork/compiler/codegen.py
        kind_map = {
            "def": SymbolKind.VARIABLE,
            "defn": SymbolKind.FUNCTION,
            "defmacro": SymbolKind.FUNCTION,
            "defclass": SymbolKind.CLASS,
            "defprotocol": SymbolKind.INTERFACE,
        }

        if head_name not in kind_map:
            return None

        name = form[1]
        if not isinstance(name, Symbol):
            return None

        # Get location from the form if available
        line = getattr(form, "line", 1) - 1  # Convert to 0-based
        col = getattr(form, "col", 0)
        end_line = getattr(form, "end_line", line + 1) - 1
        end_col = getattr(form, "end_col", col)

        return {
            "name": name.name,
            "kind": kind_map[head_name],
            "range": make_range(line, col, end_line, end_col),
            "selectionRange": make_range(line, col, line, col + len(name.name)),
        }

    # =========================================================================
    # Validation and Diagnostics
    # =========================================================================

    def _validate_document(self, doc: TextDocument) -> None:
        """Validate a document and publish diagnostics."""
        diagnostics = []

        try:
            from spork.compiler.reader import read_str

            # Try to parse the document
            read_str(doc.content)

        except Exception as e:
            # Parse error
            error_msg = str(e)

            # Try to extract line/col from error message
            line = 0
            col = 0

            # Common error patterns from spork/compiler/reader.py:
            # "unterminated list at line 3, expected )"
            # "unterminated string starting at line 5"
            # "Map literal must have even number of forms at line 10"
            import re

            # Try "at line X" pattern first (most specific)
            line_match = re.search(r"at line (\d+)", error_msg, re.IGNORECASE)
            if not line_match:
                # Try "line X" or "line: X" pattern
                line_match = re.search(r"line[:\s]+(\d+)", error_msg, re.IGNORECASE)

            col_match = re.search(r"col(?:umn)?[:\s]+(\d+)", error_msg, re.IGNORECASE)

            if line_match:
                line = int(line_match.group(1)) - 1  # Convert to 0-based
            if col_match:
                col = int(col_match.group(1))

            diagnostics.append(
                make_diagnostic(
                    range_=make_range(line, col, line, col + 1),
                    message=error_msg,
                    severity=DiagnosticSeverity.ERROR,
                    source="spork",
                )
            )

        # Try to compile and catch any compile-time errors
        if not diagnostics:
            try:
                from spork.compiler import macroexpand_all, read_str
                from spork.compiler.codegen import compile_module

                forms = read_str(doc.content)
                expanded = [macroexpand_all(f) for f in forms]
                compile_module(expanded, filename=doc.path)

            except Exception as e:
                error_msg = str(e)

                # Extract location if available
                line = 0
                col = 0

                import re

                # Try "at line X" pattern first (most specific)
                line_match = re.search(r"at line (\d+)", error_msg, re.IGNORECASE)
                if not line_match:
                    # Try "line X" or "line: X" pattern
                    line_match = re.search(r"line[:\s]+(\d+)", error_msg, re.IGNORECASE)

                col_match = re.search(
                    r"col(?:umn)?[:\s]+(\d+)", error_msg, re.IGNORECASE
                )

                if line_match:
                    line = int(line_match.group(1)) - 1
                if col_match:
                    col = int(col_match.group(1))

                diagnostics.append(
                    make_diagnostic(
                        range_=make_range(line, col, line, col + 1),
                        message=error_msg,
                        severity=DiagnosticSeverity.ERROR,
                        source="spork",
                    )
                )

        self._publish_diagnostics(doc.uri, diagnostics)

    def _publish_diagnostics(self, uri: str, diagnostics: list[dict[str, Any]]) -> None:
        """Publish diagnostics for a document."""
        self.protocol.send_notification(
            "textDocument/publishDiagnostics",
            {
                "uri": uri,
                "diagnostics": diagnostics,
            },
        )

    # =========================================================================
    # Initialization Helpers
    # =========================================================================

    def _load_project(self) -> None:
        """Try to load project configuration from spork.it."""
        if not self.root_path:
            return

        try:
            from spork.project import ProjectConfig, ProjectManager

            self.project_config = ProjectConfig.load(self.root_path)
            self.project_manager = ProjectManager(self.project_config)

            self._log(f"Loaded project: {self.project_config.name}")

            # Ensure venv exists
            if not self.project_manager.has_venv():
                self._log("Project venv not found, initializing...")
                self.project_manager.install_dependencies(quiet=True)

            # Inject venv paths
            self.project_manager.inject_venv_paths()

        except FileNotFoundError:
            self._log("No spork.it project file found")
        except Exception as e:
            self._log(f"Error loading project: {e}")

    def _init_backend(self) -> None:
        """Initialize the REPL backend for language features."""
        try:
            from spork.repl.backend import ReplBackend
            from spork.runtime.ns import init_source_roots

            # Initialize source roots
            if self.project_config:
                for source_path in self.project_config.get_absolute_source_paths():
                    if os.path.isdir(source_path):
                        init_source_roots(extra_paths=[source_path])
            else:
                init_source_roots(include_cwd=True)

            self.backend = ReplBackend()
            self._log("REPL backend initialized")

        except Exception as e:
            self._log(f"Error initializing backend: {e}")
            self.backend = None

    def _get_completion_prefix(
        self, doc: TextDocument, line: int, character: int
    ) -> str:
        """Get the completion prefix at the given position."""
        line_text = doc.get_line(line)

        if not line_text or character == 0:
            return ""

        # Find the start of the current word
        def is_symbol_char(c: str) -> bool:
            return c.isalnum() or c in "-_/.?!+*<>=:"

        start = character
        while start > 0 and is_symbol_char(line_text[start - 1]):
            start -= 1

        return line_text[start:character]

    # =========================================================================
    # Main Entry Point
    # =========================================================================

    def run(self) -> None:
        """Run the language server main loop."""
        self._log("Spork Language Server starting")

        try:
            self.protocol.run()
        except KeyboardInterrupt:
            self._log("Interrupted")
        except Exception as e:
            self._log(f"Server error: {e}")
            traceback.print_exc(file=sys.stderr)
        finally:
            self._log("Server stopped")


def start_server(log_path: Optional[str] = None) -> None:
    """
    Start the Spork language server.

    Args:
        log_path: Optional path to a log file for debugging.
    """
    log_file = None
    if log_path:
        log_file = open(log_path, "w")

    try:
        server = SporkLanguageServer()
        if log_file:
            server.log_file = log_file
        server.run()
    finally:
        if log_file:
            log_file.close()


def main() -> None:
    """Main entry point for spork-lsp command."""
    import argparse

    parser = argparse.ArgumentParser(description="Spork Language Server")
    parser.add_argument(
        "--log",
        metavar="FILE",
        help="Log file for debugging",
    )
    args = parser.parse_args()

    start_server(log_path=args.log)


if __name__ == "__main__":
    main()



"""
spork.compiler.loader - Spork import hook and compilation cache

This module handles:
- In-memory compilation cache with automatic invalidation
- Import hooks for .spork files (SporkLoader, SporkFinder)
- Build utilities for compiling .spork files to Python source

Cache invalidation happens when:
- The .spork file's mtime changes
- The compiler cache version changes (COMPILER_CACHE_VERSION)
"""

import ast
import importlib.abc
import importlib.util
import json
import os
import sys
import threading
from pathlib import Path
from types import CodeType
from typing import Any, Optional

# Version string for cache invalidation - bump when codegen changes
COMPILER_CACHE_VERSION = "spork-compiler-v1"

# Process-local compile cache
# Key: (absolute_path, mtime, COMPILER_CACHE_VERSION)
# Value: (code, macro_env)
_COMPILE_CACHE: dict[tuple[str, float, str], tuple[CodeType, dict[str, Any]]] = {}
_CACHE_LOCK = threading.Lock()


def get_cache_key(path: str) -> tuple[str, float, str]:
    """Generate a cache key for the given file path."""
    abs_path = os.path.abspath(path)
    mtime = os.path.getmtime(abs_path)
    return (abs_path, mtime, COMPILER_CACHE_VERSION)


def get_cached_code(
    path: str,
) -> Optional[tuple[CodeType, dict[str, Any]]]:
    """
    Look up compiled code in the cache.

    Returns (code, macro_env) if cached and valid, None otherwise.
    """
    try:
        key = get_cache_key(path)
    except OSError:
        return None

    with _CACHE_LOCK:
        return _COMPILE_CACHE.get(key)


def cache_compiled_code(path: str, code: CodeType, macro_env: dict[str, Any]) -> None:
    """Store compiled code in the cache."""
    try:
        key = get_cache_key(path)
    except OSError:
        return

    with _CACHE_LOCK:
        _COMPILE_CACHE[key] = (code, macro_env)


def clear_cache() -> None:
    """Clear the compilation cache (useful for testing)."""
    with _CACHE_LOCK:
        _COMPILE_CACHE.clear()


def compile_with_cache(src: str, path: str) -> tuple[CodeType, dict[str, Any]]:
    """
    Compile Spork source, using cache if available.

    Args:
        src: The source code string
        path: The file path (for cache key and error messages)

    Returns:
        (compiled_code, macro_env) tuple
    """
    # Import here to avoid circular imports
    from spork.compiler.codegen import compile_forms_to_code

    # Check cache first
    cached = get_cached_code(path)
    if cached is not None:
        return cached

    # Compile
    code, macro_env = compile_forms_to_code(src, path)

    # Cache the result
    cache_compiled_code(path, code, macro_env)

    return code, macro_env


# =============================================================================
# Import Hooks
# =============================================================================


class SporkLoader(importlib.abc.Loader):
    """
    Loader for .spork files.

    Uses the compilation cache to avoid recompiling unchanged files.
    """

    def __init__(self, fullname: str, path: str):
        self.fullname = fullname
        self.path = path

    def create_module(self, spec):
        return None  # default module creation

    def exec_module(self, module):
        from spork.runtime import setup_runtime_env

        path = os.path.abspath(self.path)

        # Try to get from cache
        cached = get_cached_code(path)

        if cached is not None:
            code, macro_env = cached
        else:
            # Need to compile
            with open(path, encoding="utf-8") as f:
                src = f.read()

            code, macro_env = compile_with_cache(src, path)

        # Setup runtime environment in module (always, not cached)
        setup_runtime_env(module.__dict__)

        # Expose this module's macro env for import-macros
        module.__dict__["__spork_macros__"] = macro_env

        # Execute the code
        exec(code, module.__dict__, module.__dict__)


class SporkFinder(importlib.abc.MetaPathFinder):
    """
    Meta path finder for .spork files.

    Searches sys.path (and provided path) for .spork files matching
    the requested module name.
    """

    def find_spec(self, fullname, path, target=None):
        # simple: only non-package modules, .spork alongside normal paths
        name = fullname.rpartition(".")[2]
        search_paths = path or sys.path
        for base in search_paths:
            candidate = os.path.join(base, name + ".spork")
            if os.path.isfile(candidate):
                loader = SporkLoader(fullname, candidate)
                return importlib.util.spec_from_loader(
                    fullname, loader, origin=candidate
                )
        return None


def install_import_hook():
    """Install the .spork file import hook."""
    for finder in sys.meta_path:
        if isinstance(finder, SporkFinder):
            return
    sys.meta_path.insert(0, SporkFinder())


# =============================================================================
# Build Utilities
# =============================================================================


def compile_file_to_python(src: str, src_path: str) -> tuple[str, dict[str, Any]]:
    """
    Compile Spork source to Python source code and generate source map.

    Args:
        src: The Spork source code
        src_path: Path to the source file (for error messages and source map)

    Returns:
        (python_source, source_map) tuple where source_map is a dict
        containing mapping information
    """
    from spork.compiler.codegen import (
        compile_defn,
        compile_module,
        get_compile_context,
        normalize_name,
    )
    from spork.compiler.macros import (
        MACRO_ENV,
        macroexpand_all,
        process_import_macros,
    )
    from spork.compiler.macros import (
        process_defmacros as _process_defmacros_base,
    )
    from spork.compiler.reader import read_str

    # Set up compilation context with filename
    ctx = get_compile_context()
    ctx.current_file = src_path if src_path != "<string>" else None

    # Phase 1: Read
    forms = read_str(src)

    # Phase 1.5: Process defmacros
    local_macro_env = dict(MACRO_ENV)
    forms = _process_defmacros_base(
        forms, local_macro_env, compile_defn, normalize_name
    )

    # Phase 1.6: Process import-macros
    forms = process_import_macros(forms, local_macro_env)

    # Phase 2: Macroexpand
    forms = macroexpand_all(forms, local_macro_env)

    # Phase 3 & 4: Analyze & Lower to AST
    mod = compile_module(forms, filename=src_path)

    # Generate Python source
    python_source = ast.unparse(mod)

    # Generate source map from AST node locations
    source_map = generate_source_map(mod, src_path)

    return python_source, source_map


def generate_source_map(mod: ast.Module, spork_file: str) -> dict[str, Any]:
    """
    Generate a source map from a Python AST module.

    Extracts line/column mappings from AST nodes that have location info.
    """
    mappings = []

    class LocationVisitor(ast.NodeVisitor):
        def generic_visit(self, node):
            if hasattr(node, "lineno") and hasattr(node, "col_offset"):
                # Get the original spork location if preserved
                # The AST node's location should reflect the spork source
                mapping = {
                    "py_line": node.lineno,
                    "py_col": node.col_offset,
                    "spork_line": node.lineno,  # Same for now, we preserve locations
                    "spork_col": node.col_offset,
                }
                if hasattr(node, "end_lineno") and node.end_lineno:
                    mapping["py_end_line"] = node.end_lineno
                    mapping["spork_end_line"] = node.end_lineno
                if hasattr(node, "end_col_offset") and node.end_col_offset:
                    mapping["py_end_col"] = node.end_col_offset
                    mapping["spork_end_col"] = node.end_col_offset
                mappings.append(mapping)
            super().generic_visit(node)

    LocationVisitor().visit(mod)

    # Deduplicate mappings (same py_line, py_col)
    seen = set()
    unique_mappings = []
    for m in mappings:
        key = (m["py_line"], m["py_col"])
        if key not in seen:
            seen.add(key)
            unique_mappings.append(m)

    return {
        "version": 1,
        "spork_file": spork_file,
        "mappings": unique_mappings,
    }


def compile_path_to_python(path: Path) -> tuple[str, dict[str, Any]]:
    """
    Compile a .spork file to Python source and source map.

    Args:
        path: Path to the .spork file

    Returns:
        (python_source, source_map) tuple
    """
    with open(path, encoding="utf-8") as f:
        src = f.read()
    return compile_file_to_python(src, str(path))



"""
Spork nREPL Server - Network REPL for editor integration.

This provides a network-based REPL server that editors and tools can connect to.
Compatible with nREPL protocol conventions.
"""

import json
import socket
import sys
import threading
import uuid
from typing import Any, Optional

from spork.repl.backend import NReplProtocol, ReplBackend


class NReplServer:
    """
    Network REPL server for editor integration.

    This server listens on a port and handles nREPL-style messages
    for code evaluation, completion, and documentation.
    """

    def __init__(self, host: str = "127.0.0.1", port: int = 7888):
        """
        Initialize the nREPL server.

        Args:
            host: The host to bind to.
            port: The port to listen on.
        """
        self.host = host
        self.port = port
        self.sessions: dict[str, ReplBackend] = {}
        self.socket = None
        self.running = False

    def create_session(self) -> str:
        """
        Create a new REPL session.

        Returns:
            The session ID.
        """
        session_id = str(uuid.uuid4())
        self.sessions[session_id] = ReplBackend()
        return session_id

    def get_or_create_session(
        self, session_id: Optional[str]
    ) -> tuple[str, ReplBackend]:
        """
        Get an existing session or create a new one.

        Args:
            session_id: Optional session ID.

        Returns:
            A tuple of (session_id, backend).
        """
        if session_id and session_id in self.sessions:
            return session_id, self.sessions[session_id]

        new_session_id = self.create_session()
        return new_session_id, self.sessions[new_session_id]

    def handle_message(self, message: dict[str, Any]) -> dict[str, Any]:
        """
        Handle an incoming nREPL message.

        Args:
            message: The message dictionary.

        Returns:
            A response dictionary.
        """
        try:
            # Debug logging - raw message
            print(f"[nREPL] RAW MESSAGE: {message}", file=sys.stderr)
            sys.stderr.flush()

            op = message.get("op")
            msg_id = message.get("id")
            session = message.get("session")

            # Debug logging
            print(
                f"[nREPL] Message: op={op}, session={session}, id={msg_id}",
                file=sys.stderr,
            )
            sys.stderr.flush()

            response = {"id": msg_id}

            if op == "clone":
                # Create a new session
                new_session = self.create_session()
                print(f"[nREPL] Created new session: {new_session}", file=sys.stderr)
                sys.stderr.flush()
                print(
                    f"[nREPL] Active sessions: {list(self.sessions.keys())}",
                    file=sys.stderr,
                )
                sys.stderr.flush()
                response["new-session"] = new_session
                response["status"] = ["done"]

            elif op == "close":
                # Close a session
                print(f"[nREPL] Closing session: {session}", file=sys.stderr)
                sys.stderr.flush()
                if session in self.sessions:
                    del self.sessions[session]
                response["status"] = ["done", "session-closed"]

            elif op == "eval":
                # Evaluate code
                code = message.get("code", "")
                file_path = message.get("file", None)  # Source file path
                ns = message.get("ns", None)  # Namespace context
                print(f"[nREPL] Requested session: {session}", file=sys.stderr)
                sys.stderr.flush()
                print(
                    f"[nREPL] Available sessions: {list(self.sessions.keys())}",
                    file=sys.stderr,
                )
                sys.stderr.flush()
                session_id, backend = self.get_or_create_session(session)

                # Debug logging
                print(
                    f"[nREPL] Using session {session_id} (same as requested: {session_id == session})",
                    file=sys.stderr,
                )
                sys.stderr.flush()
                print(f"[nREPL] Eval code: {code[:80]}...", file=sys.stderr)
                if file_path:
                    print(f"[nREPL] File: {file_path}", file=sys.stderr)
                if ns:
                    print(f"[nREPL] Namespace: {ns}", file=sys.stderr)
                sys.stderr.flush()
                env_keys = [
                    k for k in backend.state.env.keys() if not k.startswith("__")
                ]
                print(
                    f"[nREPL] Backend env has {len(backend.state.env)} keys, non-__ keys: {env_keys[:10]}",
                    file=sys.stderr,
                )
                print(
                    f"[nREPL] Backend ID: {id(backend)}, State ID: {id(backend.state)}, Env ID: {id(backend.state.env)}",
                    file=sys.stderr,
                )
                sys.stderr.flush()

                protocol = NReplProtocol(backend)
                result = protocol.handle_eval(
                    code,
                    session_id,
                    file_path=file_path,
                    ns=ns,  # type: ignore[call-arg]
                )

                # Debug logging after eval
                env_keys_after = [
                    k for k in backend.state.env.keys() if not k.startswith("__")
                ]
                print(
                    f"[nREPL] After eval, non-__ env keys: {env_keys_after[:10]}",
                    file=sys.stderr,
                )
                print(
                    f"[nREPL] After eval - Backend ID: {id(backend)}, State ID: {id(backend.state)}, Env ID: {id(backend.state.env)}",
                    file=sys.stderr,
                )
                sys.stderr.flush()

                response.update(result)
                # Include current namespace in response
                response["ns"] = backend.state.namespace

            elif op == "load-file":
                # Load a file
                file_content = message.get("file", "")
                file_path = message.get("file-path", "<loaded-file>")

                session_id, backend = self.get_or_create_session(session)
                protocol = NReplProtocol(backend)

                result = protocol.handle_eval(
                    file_content,
                    session_id,
                    file_path=file_path,  # type: ignore[call-arg]
                )
                response.update(result)
                # Include current namespace in response
                response["ns"] = backend.state.namespace

            elif op == "complete":
                # Auto-completion
                prefix = message.get("prefix", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_complete(prefix)
                response.update(result)

            elif op == "info":
                # Get symbol info (rich metadata)
                symbol = message.get("symbol", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_info(symbol)
                response.update(result)

            elif op == "macroexpand":
                # Macroexpand code
                code = message.get("code", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_macroexpand(code)
                response.update(result)

            elif op == "transpile":
                # Transpile Spork code to Python
                code = message.get("code", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_transpile(code)
                response.update(result)

            elif op == "find-def":
                # Find definition location
                symbol = message.get("symbol", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_find_def(symbol)
                response.update(result)

            elif op == "inspect-start":
                # Start inspector session
                code = message.get("code", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_inspect_start(code)
                response.update(result)

            elif op == "inspect-nav":
                # Navigate in inspector
                handle = message.get("handle", 0)
                path = message.get("path", [])
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_inspect_nav(handle, path)
                response.update(result)

            elif op == "protocols":
                # Get all registered protocols
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_protocols()
                response.update(result)

            elif op == "using-ns":
                # Switch to a namespace
                ns_name = message.get("ns", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_using_ns(ns_name)
                response.update(result)

            elif op == "ns-list":
                # List all loaded namespaces
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_ns_list()
                response.update(result)

            elif op == "ns-info":
                # Get info about a namespace
                ns_name = message.get("ns", "")
                session_id, backend = self.get_or_create_session(session)

                protocol = NReplProtocol(backend)
                result = protocol.handle_ns_info(ns_name)
                response.update(result)

            elif op == "describe":
                # Describe the server
                response["versions"] = {
                    "spork": {"version-string": "0.1.0"},
                    "python": {"version-string": "3.x"},
                }
                response["ops"] = {
                    "clone": {},
                    "close": {},
                    "eval": {},
                    "load-file": {},
                    "complete": {},
                    "info": {},
                    "describe": {},
                    "macroexpand": {},
                    "find-def": {},
                    "inspect-start": {},
                    "inspect-nav": {},
                    "protocols": {},
                    "using-ns": {},
                    "ns-list": {},
                    "ns-info": {},
                }
                response["status"] = ["done"]

            else:
                response["status"] = ["error", "unknown-op"]
                response["error"] = f"Unknown operation: {op}"

            print(f"[nREPL] Returning response: {response}", file=sys.stderr)
            sys.stderr.flush()
            return response

        except Exception as e:
            print(f"[nREPL] EXCEPTION in handle_message: {e}", file=sys.stderr)
            import traceback

            traceback.print_exc(file=sys.stderr)
            sys.stderr.flush()
            return {
                "id": message.get("id"),
                "status": ["error"],
                "error": str(e),
                "traceback": traceback.format_exc(),
            }

    def handle_client(self, client_socket: socket.socket, addr):
        """
        Handle a client connection.

        Args:
            client_socket: The client socket.
            addr: The client address.
        """
        print(f"Client connected from {addr}")
        buffer = b""

        try:
            while self.running:
                data = client_socket.recv(4096)
                if not data:
                    break

                buffer += data

                # Process complete messages (newline-delimited JSON)
                while b"\n" in buffer:
                    line, buffer = buffer.split(b"\n", 1)

                    if not line.strip():
                        continue

                    try:
                        message = json.loads(line.decode("utf-8"))
                        response = self.handle_message(message)

                        # Send response as JSON
                        response_json = json.dumps(response) + "\n"
                        client_socket.sendall(response_json.encode("utf-8"))

                    except json.JSONDecodeError as e:
                        error_response = {
                            "status": ["error"],
                            "error": f"Invalid JSON: {e}",
                        }
                        client_socket.sendall(
                            (json.dumps(error_response) + "\n").encode("utf-8")
                        )
                    except Exception as e:
                        error_response = {
                            "status": ["error"],
                            "error": str(e),
                        }
                        client_socket.sendall(
                            (json.dumps(error_response) + "\n").encode("utf-8")
                        )

        except Exception as e:
            print(f"Error handling client {addr}: {e}")
        finally:
            client_socket.close()
            print(f"Client disconnected from {addr}")

    def start(self):
        """Start the nREPL server."""
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((self.host, self.port))
        self.socket.listen(5)
        self.running = True

        print(f"Spork nREPL server started on {self.host}:{self.port}")

        # Write port file for editors
        with open(".nrepl-port", "w") as f:
            f.write(str(self.port))

        try:
            while self.running:
                try:
                    client_socket, addr = self.socket.accept()
                    client_thread = threading.Thread(
                        target=self.handle_client, args=(client_socket, addr)
                    )
                    client_thread.daemon = True
                    client_thread.start()
                except OSError:
                    if not self.running:
                        break
                    raise
        except KeyboardInterrupt:
            print("\nShutting down...")
        finally:
            self.stop()

    def stop(self):
        """Stop the nREPL server."""
        self.running = False
        if self.socket:
            self.socket.close()
        print("Server stopped.")


class SimpleNReplClient:
    """
    A simple nREPL client for testing and scripting.
    """

    def __init__(self, host: str = "127.0.0.1", port: int = 7888):
        """
        Initialize the client.

        Args:
            host: The host to connect to.
            port: The port to connect to.
        """
        self.host = host
        self.port = port
        self.socket = None
        self.session = None
        self.msg_counter = 0

    def connect(self):
        """Connect to the nREPL server."""
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.host, self.port))

        # Create a session
        response = self.send_message({"op": "clone"})
        self.session = response.get("new-session")
        print(f"Connected to {self.host}:{self.port}")
        print(f"Session: {self.session}")

    def send_message(self, message: dict[str, Any]) -> dict[str, Any]:
        """
        Send a message and receive the response.

        Args:
            message: The message to send.

        Returns:
            The response dictionary.
        """
        if not self.socket:
            raise RuntimeError("Not connected")

        # Add message ID
        self.msg_counter += 1
        message["id"] = str(self.msg_counter)

        # Add session if we have one
        if self.session and "session" not in message:
            message["session"] = self.session

        # Send message
        message_json = json.dumps(message) + "\n"
        self.socket.sendall(message_json.encode("utf-8"))

        # Receive response
        buffer = b""
        while True:
            data = self.socket.recv(4096)
            if not data:
                raise RuntimeError("Connection closed")

            buffer += data
            if b"\n" in buffer:
                line, buffer = buffer.split(b"\n", 1)
                return json.loads(line.decode("utf-8"))

    def eval(self, code: str) -> Any:
        """
        Evaluate code on the server.

        Args:
            code: The code to evaluate.

        Returns:
            The evaluation result.
        """
        response = self.send_message({"op": "eval", "code": code})
        return response

    def complete(self, prefix: str) -> list[str]:
        """
        Get completions for a prefix.

        Args:
            prefix: The prefix to complete.

        Returns:
            A list of completions.
        """
        response = self.send_message({"op": "complete", "prefix": prefix})
        return response.get("completions", [])

    def close(self):
        """Close the connection."""
        if self.socket:
            try:
                self.send_message({"op": "close"})
            except Exception:
                pass
            self.socket.close()
            self.socket = None
            print("Disconnected.")


def main():
    """Main entry point for the nREPL server."""
    import argparse

    parser = argparse.ArgumentParser(description="Spork nREPL Server")
    parser.add_argument("--host", default="127.0.0.1", help="Host to bind to")
    parser.add_argument("--port", type=int, default=7888, help="Port to listen on")
    parser.add_argument(
        "--client", action="store_true", help="Run as a test client instead"
    )

    args = parser.parse_args()

    if args.client:
        # Run as client
        client = SimpleNReplClient(args.host, args.port)
        try:
            client.connect()

            print("\nSimple nREPL Client")
            print("Type code to evaluate, or :quit to exit\n")

            while True:
                try:
                    code = input("client> ")
                    if code.strip() == ":quit":
                        break
                    if not code.strip():
                        continue

                    response = client.eval(code)

                    if "value" in response:
                        print(f"=> {response['value']}")
                    elif "error" in response:
                        print(f"Error: {response['error']}")

                except EOFError:
                    break
                except KeyboardInterrupt:
                    print()
                    continue
                except Exception as e:
                    print(f"Error: {e}")
        finally:
            client.close()
    else:
        # Run as server
        server = NReplServer(args.host, args.port)
        server.start()


if __name__ == "__main__":
    main()



"""
Spork Namespace System - Manages namespace resolution, registration, and lookup.

This module provides the infrastructure for Spork's namespace system, including:
- NAMESPACE_REGISTRY: Global registry of loaded namespaces
- SOURCE_ROOTS: Ordered list of directories to search for .spork files
- Resolution functions to map namespace names to file paths
- Registration functions to track loaded namespaces
"""

import os
from dataclasses import dataclass, field
from typing import Any, Callable, Optional


class NamespaceProxy:
    """
    A proxy object that wraps a namespace's environment dict.

    This allows attribute-style access to namespace symbols, e.g.:
        h.add_nums  instead of  h["add_nums"]

    Also handles hyphen-to-underscore normalization for Spork names.
    """

    def __init__(self, env: dict[str, Any], ns_name: str):
        object.__setattr__(self, "_env", env)
        object.__setattr__(self, "_ns_name", ns_name)

    def __getattr__(self, name: str) -> Any:
        env = object.__getattribute__(self, "_env")
        ns_name = object.__getattribute__(self, "_ns_name")

        # Try the name as-is first
        if name in env:
            return env[name]

        # Try with hyphens converted to underscores (Spork convention)
        hyphen_name = name.replace("_", "-")
        if hyphen_name in env:
            return env[hyphen_name]

        # Try underscore version
        underscore_name = name.replace("-", "_")
        if underscore_name in env:
            return env[underscore_name]

        raise AttributeError(f"Namespace '{ns_name}' has no symbol '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        env = object.__getattribute__(self, "_env")
        env[name] = value

    def __repr__(self) -> str:
        ns_name = object.__getattribute__(self, "_ns_name")
        return f"<NamespaceProxy '{ns_name}'>"

    def __dir__(self) -> list[str]:
        env = object.__getattribute__(self, "_env")
        return [k for k in env.keys() if not k.startswith("_")]


@dataclass
class NamespaceInfo:
    """Information about a loaded namespace."""

    name: str  # Fully qualified namespace name, e.g., "my.app.core"
    file: Optional[str]  # Absolute path to the .spork file, or None for built-in
    env: dict[str, Any]  # The namespace's environment dict
    macros: dict[str, Callable]  # Macros defined in this namespace
    loaded: bool = True
    timestamp: float = 0.0  # Last modification time of file when loaded
    refers: dict[str, str] = field(
        default_factory=dict
    )  # symbol -> source-ns for referred symbols
    aliases: dict[str, str] = field(
        default_factory=dict
    )  # alias -> full-ns-name for :as aliases


# Global namespace registry
# Maps namespace name (str) -> NamespaceInfo
NAMESPACE_REGISTRY: dict[str, NamespaceInfo] = {}

# Ordered list of source roots to search for .spork files
# Populated at startup from:
# 1. Directory of current file being executed
# 2. Project root (if detected)
# 3. Current working directory
# 4. $SPORK_PATH entries
# 5. CLI --source-path flags
SOURCE_ROOTS: list[str] = []


def ns_to_relpath(ns: str) -> str:
    """
    Convert a namespace name to a relative file path.

    Examples:
        "my.app.core" -> "my/app/core.spork"
        "spork.pds" -> "spork/pds.spork"
        "utils" -> "utils.spork"
    """
    return ns.replace(".", os.sep) + ".spork"


def relpath_to_ns(relpath: str) -> str:
    """
    Convert a relative file path to a namespace name.

    Examples:
        "my/app/core.spork" -> "my.app.core"
        "utils.spork" -> "utils"
    """
    # Remove .spork extension
    if relpath.endswith(".spork"):
        relpath = relpath[:-6]
    # Convert path separators to dots
    return relpath.replace(os.sep, ".").replace("/", ".")


def _get_std_dir() -> str:
    """Get the path to the spork/std directory."""
    return os.path.join(os.path.dirname(os.path.dirname(__file__)), "std")


def find_spork_file_for_ns(
    ns: str, extra_roots: Optional[list[str]] = None
) -> Optional[str]:
    """
    Find the .spork file for a given namespace name.

    Searches through SOURCE_ROOTS (and any extra_roots) in order,
    looking for the corresponding .spork file.

    For std.* namespaces, also searches the spork package's std/ directory.

    Args:
        ns: The namespace name, e.g., "my.app.core"
        extra_roots: Additional directories to search (searched first)

    Returns:
        Absolute path to the .spork file if found, None otherwise.
    """
    # Handle std.* namespaces specially - look in spork/std/
    if ns.startswith("std."):
        # Convert std.string -> string.spork in std/ dir
        std_rel = ns[4:].replace(".", os.sep) + ".spork"
        std_path = os.path.join(_get_std_dir(), std_rel)
        if os.path.isfile(std_path):
            return os.path.abspath(std_path)

    rel = ns_to_relpath(ns)

    # Build search path: extra_roots first, then SOURCE_ROOTS
    search_paths = []
    if extra_roots:
        search_paths.extend(extra_roots)
    search_paths.extend(SOURCE_ROOTS)

    for root in search_paths:
        if not root:
            continue
        candidate = os.path.join(root, rel)
        if os.path.isfile(candidate):
            return os.path.abspath(candidate)

    return None


def find_project_root(start_path: str) -> Optional[str]:
    """
    Find the project root by walking up from start_path.

    Looks for common project markers:
    - .git directory
    - pyproject.toml
    - .spork_project file
    - setup.py

    Args:
        start_path: Path to start searching from (usually a file path)

    Returns:
        Absolute path to project root if found, None otherwise.
    """
    if os.path.isfile(start_path):
        current = os.path.dirname(os.path.abspath(start_path))
    else:
        current = os.path.abspath(start_path)

    markers = [".git", "pyproject.toml", ".spork_project", "setup.py"]

    while current != os.path.dirname(current):  # Stop at filesystem root
        for marker in markers:
            if os.path.exists(os.path.join(current, marker)):
                return current
        current = os.path.dirname(current)

    return None


def init_source_roots(
    current_file: Optional[str] = None,
    extra_paths: Optional[list[str]] = None,
    include_cwd: bool = True,
) -> None:
    """
    Initialize SOURCE_ROOTS based on context.

    This should be called at startup (e.g., in exec_file or REPL init).

    Args:
        current_file: The file being executed (if any)
        extra_paths: Additional paths from CLI --source-path flags
        include_cwd: Whether to include current working directory
    """
    global SOURCE_ROOTS
    SOURCE_ROOTS = []

    # 1. Directory of current file
    if current_file:
        file_dir = os.path.dirname(os.path.abspath(current_file))
        if file_dir and file_dir not in SOURCE_ROOTS:
            SOURCE_ROOTS.append(file_dir)

        # 2. Project root (if detected)
        project_root = find_project_root(current_file)
        if project_root and project_root not in SOURCE_ROOTS:
            SOURCE_ROOTS.append(project_root)

    # 3. Current working directory
    if include_cwd:
        cwd = os.getcwd()
        if cwd not in SOURCE_ROOTS:
            SOURCE_ROOTS.append(cwd)

    # 4. $SPORK_PATH entries
    spork_path = os.environ.get("SPORK_PATH", "")
    if spork_path:
        for p in spork_path.split(os.pathsep):
            p = p.strip()
            if p and os.path.isdir(p) and p not in SOURCE_ROOTS:
                SOURCE_ROOTS.append(p)

    # 5. CLI --source-path flags
    if extra_paths:
        for p in extra_paths:
            p = os.path.abspath(p)
            if p not in SOURCE_ROOTS:
                SOURCE_ROOTS.append(p)


def add_source_root(path: str, prepend: bool = False) -> None:
    """
    Add a directory to SOURCE_ROOTS.

    Args:
        path: Directory path to add
        prepend: If True, add to front of list (higher priority)
    """
    path = os.path.abspath(path)
    if path not in SOURCE_ROOTS:
        if prepend:
            SOURCE_ROOTS.insert(0, path)
        else:
            SOURCE_ROOTS.append(path)


def register_namespace(
    name: str,
    file: Optional[str],
    env: dict[str, Any],
    macros: dict[str, Callable],
    refers: Optional[dict[str, str]] = None,
    aliases: Optional[dict[str, str]] = None,
) -> NamespaceInfo:
    """
    Register a namespace in the global registry.

    Args:
        name: Fully qualified namespace name
        file: Absolute path to the source file (or None for synthetic/REPL ns)
        env: The namespace's environment dictionary
        macros: Macros defined in this namespace
        refers: Symbol -> source namespace mapping for referred symbols
        aliases: Alias -> namespace name mapping for :as aliases

    Returns:
        The NamespaceInfo that was registered.
    """
    timestamp = 0.0
    if file and os.path.isfile(file):
        timestamp = os.path.getmtime(file)

    info = NamespaceInfo(
        name=name,
        file=file,
        env=env,
        macros=macros,
        loaded=True,
        timestamp=timestamp,
        refers=refers or {},
        aliases=aliases or {},
    )
    NAMESPACE_REGISTRY[name] = info
    return info


def get_namespace(name: str) -> Optional[NamespaceInfo]:
    """
    Get a namespace from the registry.

    Args:
        name: Fully qualified namespace name

    Returns:
        NamespaceInfo if found, None otherwise.
    """
    return NAMESPACE_REGISTRY.get(name)


def namespace_loaded(name: str) -> bool:
    """
    Check if a namespace is already loaded.

    Args:
        name: Fully qualified namespace name

    Returns:
        True if namespace is in registry and marked as loaded.
    """
    info = NAMESPACE_REGISTRY.get(name)
    return info is not None and info.loaded


def needs_reload(name: str) -> bool:
    """
    Check if a namespace needs to be reloaded (file changed since load).

    Args:
        name: Fully qualified namespace name

    Returns:
        True if the source file has been modified since the namespace was loaded.
    """
    info = NAMESPACE_REGISTRY.get(name)
    if info is None or info.file is None:
        return False

    if not os.path.isfile(info.file):
        return False

    current_mtime = os.path.getmtime(info.file)
    return current_mtime > info.timestamp


def unload_namespace(name: str) -> bool:
    """
    Remove a namespace from the registry.

    Args:
        name: Fully qualified namespace name

    Returns:
        True if namespace was removed, False if it wasn't in registry.
    """
    if name in NAMESPACE_REGISTRY:
        del NAMESPACE_REGISTRY[name]
        return True
    return False


def list_namespaces() -> list[str]:
    """
    List all registered namespace names.

    Returns:
        List of fully qualified namespace names.
    """
    return list(NAMESPACE_REGISTRY.keys())


def clear_registry() -> None:
    """Clear all namespaces from the registry."""
    NAMESPACE_REGISTRY.clear()


# ============================================================================
# Require Resolution
# ============================================================================


def is_python_module(name: str) -> bool:
    """
    Check if a name refers to an importable Python module.

    Args:
        name: Module name (e.g., "math", "collections.abc")

    Returns:
        True if the module can be imported as a Python module.
    """
    import importlib.util

    try:
        spec = importlib.util.find_spec(name)
        return spec is not None
    except (ModuleNotFoundError, ValueError):
        return False


def resolve_require(
    ns_name: str, current_file: Optional[str] = None
) -> tuple[str, Optional[str]]:
    """
    Resolve a require target to either a Spork file or Python module.

    Args:
        ns_name: The namespace/module name to require
        current_file: The file doing the requiring (for relative resolution)

    Returns:
        A tuple of (type, path) where:
        - type is "spork" or "python"
        - path is the absolute file path for Spork, or None for Python
    """
    # Build extra roots from current file
    extra_roots = []
    if current_file:
        file_dir = os.path.dirname(os.path.abspath(current_file))
        if file_dir:
            extra_roots.append(file_dir)

    # Try to find a Spork file first
    spork_file = find_spork_file_for_ns(ns_name, extra_roots=extra_roots)
    if spork_file:
        return ("spork", spork_file)

    # Fall back to Python module
    if is_python_module(ns_name):
        return ("python", None)

    # Not found
    raise FileNotFoundError(
        f"Cannot resolve namespace '{ns_name}': "
        f"no .spork file found and not a Python module"
    )


def parse_require_spec(spec) -> dict[str, Any]:
    """
    Parse a require specification from Spork syntax.

    Handles forms like:
    - ns-name (just the symbol)
    - [ns-name]
    - [ns-name :as alias]
    - [ns-name :refer [sym1 sym2]]
    - [ns-name :as alias :refer [sym1 sym2]]
    - [ns-name :refer :all]

    Args:
        spec: A Symbol or Vector from Spork reader

    Returns:
        Dict with keys:
        - "ns": namespace name (str)
        - "alias": alias name or None
        - "refer": list of symbol names, ":all", or None
    """
    # Import here to avoid circular imports
    from spork.compiler import Keyword, Symbol, VectorLiteral

    result: dict[str, Any] = {"ns": None, "alias": None, "refer": None}

    # Handle bare symbol
    if isinstance(spec, Symbol):
        result["ns"] = spec.name
        return result

    # Handle vector form
    if isinstance(spec, VectorLiteral):
        items = spec.items
        if not items:
            raise SyntaxError("Empty require spec")

        # First element must be the namespace
        if not isinstance(items[0], Symbol):
            raise SyntaxError("Require spec must start with namespace symbol")
        result["ns"] = items[0].name

        # Parse keyword arguments
        i = 1
        while i < len(items):
            item = items[i]
            if isinstance(item, Keyword):
                if item.name == "as":
                    if i + 1 >= len(items):
                        raise SyntaxError(":as requires an alias symbol")
                    alias = items[i + 1]
                    if not isinstance(alias, Symbol):
                        raise SyntaxError(":as alias must be a symbol")
                    result["alias"] = alias.name
                    i += 2
                elif item.name == "refer":
                    if i + 1 >= len(items):
                        raise SyntaxError(":refer requires symbols or :all")
                    refer_spec = items[i + 1]
                    if isinstance(refer_spec, Keyword) and refer_spec.name == "all":
                        result["refer"] = ":all"
                    elif isinstance(refer_spec, VectorLiteral):
                        result["refer"] = []
                        for sym in refer_spec.items:
                            if not isinstance(sym, Symbol):
                                raise SyntaxError(":refer items must be symbols")
                            result["refer"].append(sym.name)
                    else:
                        raise SyntaxError(":refer must be followed by vector or :all")
                    i += 2
                else:
                    raise SyntaxError(f"Unknown require option :{item.name}")
            else:
                raise SyntaxError(f"Unexpected item in require spec: {item}")

        return result

    raise SyntaxError(f"Invalid require spec: {spec}")


def validate_ns_name(ns_name: str, file_path: Optional[str] = None) -> bool:
    """
    Validate that a namespace name matches its file path.

    Args:
        ns_name: The namespace name declared in (ns ...)
        file_path: The file containing the namespace declaration

    Returns:
        True if valid, raises error if not.
    """
    if not file_path:
        return True  # Can't validate without file path

    # Get expected namespace from file path
    # We need to figure out which source root the file is under
    abs_path = os.path.abspath(file_path)

    for root in SOURCE_ROOTS:
        root = os.path.abspath(root)
        if abs_path.startswith(root + os.sep):
            rel_path = abs_path[len(root) + 1 :]
            expected_ns = relpath_to_ns(rel_path)
            if ns_name != expected_ns:
                raise ValueError(
                    f"Namespace name '{ns_name}' does not match file path. "
                    f"Expected '{expected_ns}' based on path '{file_path}'"
                )
            return True

    # File not under any source root - can't validate, allow it
    return True



"""
Spork - A Lisp to Python transpiler with persistent data structures.
"""

import sys

# Import persistent data structures from runtime (wraps C extension)
try:
    from spork.runtime import (
        _MISSING,
        EMPTY_DOUBLE_VECTOR,
        EMPTY_LONG_VECTOR,
        EMPTY_MAP,
        EMPTY_SET,
        # Empty constants
        EMPTY_VECTOR,
        Cons,
        Decorated,
        DoubleVector,
        IntVector,
        Keyword,
        Map,
        MapLiteral,
        MatchError,
        Set,
        SetLiteral,
        # AST/Reader types
        Symbol,
        TransientDoubleVector,
        TransientIntVector,
        TransientMap,
        TransientSet,
        TransientVector,
        # PDS types
        Vector,
        VectorLiteral,
        add,
        assoc,
        assoc_bang,
        bit_and,
        bit_and_not,
        bit_not,
        # Bitwise
        bit_or,
        bit_shift_left,
        bit_shift_right,
        bit_xor,
        concat,
        conj,
        conj_bang,
        cons,
        contains_q,
        count,
        cycle,
        dec,
        dedupe,
        disj,
        disj_bang,
        dissoc,
        dissoc_bang,
        distinct,
        div,
        doall,
        dorun,
        drop,
        drop_while,
        empty,
        even_q,
        every,
        # Sequence operations
        first,
        flatten,
        frequencies,
        get,
        group_by,
        hash_map,
        hash_set,
        # Math
        inc,
        interleave,
        interpose,
        into,
        iterate,
        keep,
        keep_indexed,
        map_indexed,
        mapcat,
        mod,
        mul,
        neg_q,
        not_any,
        not_every,
        nth,
        odd_q,
        partition,
        partition_all,
        persistent_bang,
        pop_bang,
        pos_q,
        quot,
        realized_q,
        reduce,
        reductions,
        rest,
        reverse,
        seq,
        # Runtime utilities
        setup_runtime_env,
        # Predicates and reducers
        some,
        sort,
        sort_by,
        split_at,
        split_with,
        spork_abs,
        spork_filter,
        # Lazy sequences
        spork_map,
        spork_max,
        spork_min,
        spork_raise,
        spork_range,
        spork_repeat,
        spork_setattr,
        spork_try,
        sub,
        take,
        take_while,
        # Transients
        transient,
        # Constructors
        vec,
        vec_f64,
        vec_i64,
        zero_q,
        # Collection utilities
        zipmap,
    )
except ImportError as e:
    print("Error: Spork runtime could not be loaded.", file=sys.stderr)
    print(f"  {e}", file=sys.stderr)
    print("", file=sys.stderr)
    print("The C extension (pds) may not be built. Try:", file=sys.stderr)
    print("  make build", file=sys.stderr)
    print("", file=sys.stderr)
    sys.exit(1)

# Import compiler functions
from spork.compiler import (
    MACRO_ENV,
    Reader,
    compile_forms_to_code,
    eval_str,
    exec_file,
    export_file,
    install_import_hook,
    # Macros
    macroexpand,
    macroexpand_all,
    # Reader
    read_str,
    tokenize,
)

install_import_hook()

__version__ = "0.1.0"

__all__ = [
    # Data structures
    "Vector",
    "Map",
    "Set",
    "DoubleVector",
    "IntVector",
    "Cons",
    "TransientVector",
    "TransientDoubleVector",
    "TransientIntVector",
    "TransientMap",
    "TransientSet",
    "EMPTY_VECTOR",
    "EMPTY_DOUBLE_VECTOR",
    "EMPTY_LONG_VECTOR",
    "EMPTY_MAP",
    "EMPTY_SET",
    "cons",
    "vec",
    "vec_f64",
    "vec_i64",
    "hash_map",
    "hash_set",
    # AST/Reader types
    "Symbol",
    "Keyword",
    "VectorLiteral",
    "MapLiteral",
    "SetLiteral",
    "Decorated",
    "MatchError",
    "_MISSING",
    # Compilation
    "compile_forms_to_code",
    "eval_str",
    "exec_file",
    "export_file",
    "install_import_hook",
    "read_str",
    "tokenize",
    "Reader",
    "macroexpand",
    "macroexpand_all",
    "MACRO_ENV",
    # Sequence operations
    "first",
    "rest",
    "seq",
    "nth",
    "conj",
    "assoc",
    "dissoc",
    "disj",
    "get",
    "count",
    "contains_q",
    "drop",
    "concat",
    "into",
    "empty",
    "spork_map",
    "spork_filter",
    "take",
    "take_while",
    "drop_while",
    "spork_repeat",
    "cycle",
    "iterate",
    "spork_range",
    "interleave",
    "interpose",
    "partition",
    "partition_all",
    "keep",
    "keep_indexed",
    "map_indexed",
    "dedupe",
    "distinct",
    "flatten",
    "mapcat",
    "some",
    "every",
    "not_every",
    "not_any",
    "reduce",
    "reductions",
    "zipmap",
    "group_by",
    "frequencies",
    "reverse",
    "sort",
    "sort_by",
    "split_at",
    "split_with",
    "doall",
    "dorun",
    "realized_q",
    # Math
    "inc",
    "dec",
    "even_q",
    "odd_q",
    "pos_q",
    "neg_q",
    "zero_q",
    "add",
    "sub",
    "mul",
    "div",
    "mod",
    "quot",
    "spork_max",
    "spork_min",
    "spork_abs",
    # Bitwise
    "bit_or",
    "bit_and",
    "bit_and_not",
    "bit_xor",
    "bit_not",
    "bit_shift_left",
    "bit_shift_right",
    # Transients
    "transient",
    "persistent_bang",
    "conj_bang",
    "assoc_bang",
    "dissoc_bang",
    "disj_bang",
    "pop_bang",
    # Runtime utilities
    "setup_runtime_env",
    "spork_try",
    "spork_raise",
    "spork_setattr",
]
